/*
 * config.c - Reads XML configuration files.
 *
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/**
 * \file config.c
 * Reads XML configuration files. Fills a #config structure with configuration
 * data.\n Example configuration file:
 * \code
 * <config>
 *     <common 
 *          refresh="200"
 *          key_repeat_delay="65"
 *          auth_file="/etc/veronica.auth"
 *     />
 *     <screen
 *          theme_dir="/opt/veronica/themes/black_coverart"
 *          show_cursor="1"
 *          vnc_port="5900"
 *          http_port="5800"
 *          name="default"
 *     />
 *     <screen
 *          theme_dir="/opt/veronica/themes/pda_240_300"
 *          show_cursor="0"
 *          mouse_control="0"
 *          vnc_port="5901"
 *          name="pda"
 *     />
 * </config>
 * \endcode
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/xmlreader.h>
#include <rfb/rfb.h>
#include "veronica.h"
#include "xmlutil.h"
#include "config.h"

/**
 * Default configuration file.
 */
#define CONFIG_DEFAULT_FILE "/etc/veronica_conf.xml"

/**
 * Default display refresh time (milliseconds).
 */
#define CONFIG_DEFAULT_REFRESH 200

/**
 * Default delay between key presses in key repeat mode (milliseconds).
 */
#define CONFIG_DEFAULT_KEY_DELAY 65

/**
 * Default VNC port. 
 */
#define CONFIG_DEFAULT_PORT 5900

/**
 * Default setting for displaying a mouse cursor on server side.\n
 * 1 = show mouse cursor, 0 = no mouse cursor
 */
#define CONFIG_DEFAULT_SHOW_CURSOR 0

/**
 * Default setting for mouse control within display area.\n
 * 1 = mouse control, 0 = no mouse control
 */
#define CONFIG_DEFAULT_MOUSE_CONTROL 1

/**
 * Default screen name. 
 */
#define CONFIG_DEFAULT_SCREEN_NAME "Veronica"

/**
 * XML attribute for display refresh time.
 */
#define CONFIG_ATTR_REFRESH "refresh"

/**
 * XML attribute for key repeat delay.
 */
#define CONFIG_ATTR_KEY_REPEAT_DELAY "key_repeat_delay"

/**
 * XML attribute for authentication file.
 */
#define CONFIG_ATTR_AUTH_FILE "auth_file"

/**
 * XML attribute for theme directroy.
 */
#define CONFIG_ATTR_THEME_DIR "theme_dir"

/**
 * XML attribute for displaying a mouse cursor on server side.
 */
#define CONFIG_ATTR_SHOW_CURSOR "show_cursor"

/**
 * XML attribute for mouse control within display area.
 */
#define CONFIG_ATTR_MOUSE_CONTROL "mouse_control"

/**
 * XML attribute for vnc port (Remote Framebuffer Protocol).
 */
#define CONFIG_ATTR_VNC_PORT "vnc_port"

/**
 * XML attribute for http port (Java viewer applet).
 */
#define CONFIG_ATTR_HTTP_PORT "http_port"

/**
 * XML attribute for VNC screen name (displayed in the title bar).
 */
#define CONFIG_ATTR_SCREEN_NAME "name"

/**
 * XML element for the 'common' part of a configuration file.
 */
#define CONFIG_ELEM_COMMON "common"

/**
 * XML element for the 'screen' parts of a configuration file.
 */
#define CONFIG_ELEM_SERVER "screen"

/**
 * Minimum port number of listener port.
 */
#define CONFIG_MIN_PORT 0

/**
 * Maximum port number of listener port.
 */
#define CONFIG_MAX_PORT 65535

/**
 * Provides the default name for a VNC screen (displayed in the title bar of
 * VNC viewers.)
 * @return the default screen name
 */
static char *config_get_screen_name() {
	char *screen_name = NULL;
	
	screen_name = malloc(strlen(CONFIG_DEFAULT_SCREEN_NAME)
			+ strlen(VERONICA_VERSION) + 2);
	if (screen_name == NULL) {
		rfbLogPerror("config_get_screen_name: malloc");
		return NULL;
	} 
	sprintf(screen_name, "%s %s", CONFIG_DEFAULT_SCREEN_NAME,
			VERONICA_VERSION);

	return screen_name;
}

/**
 * Processes the common node in a configuration XML document.
 * @param reader  the XML reader
 * @param config  configuration data
 * @return 0 on success or -1 if an error occured
 */
static int config_xml_process_common(xmlTextReaderPtr reader, 
		struct config* config)
{
	const xmlChar *name, *value;
	int refresh = 0, key_repeat_delay = 0;
	char *auth_file = NULL;
	int line = xmlutil_get_current_line(reader);

	while (xmlTextReaderMoveToNextAttribute(reader) > 0) {
		if (xmlutil_read_attribute(reader, line, &name, &value) < 0) {
			return -1;
		}
		if (strcasecmp(name, CONFIG_ATTR_REFRESH) == 0) {
			config->refresh = atoi(value);
			rfbLog("Common - refresh: %d ms\n", config->refresh);
		}
		else if (strcasecmp(name, CONFIG_ATTR_KEY_REPEAT_DELAY) == 0) {
			config->key_repeat_delay = atoi(value);
			rfbLog("Common - key_repeat_delay: %d ms\n",
					config->key_repeat_delay);
		}
		else if (strcasecmp(name, CONFIG_ATTR_AUTH_FILE) == 0) {
			if (strlen(value) > 0) {
				if ((config->auth_file = strdup(value)) == NULL) {
					rfbLogPerror("config_xml_process_common: strdup");
					return -1;
				}
				rfbLog("Common - auth_file: %s\n", config->auth_file);
			}
		}
	}

	return 0;
}

/**
 * Processes a screen node in a configuration XML document.
 * @param reader  the XML reader
 * @param config  configuration data
 * @return 0 on success or -1 if an error occured
 */
static int config_xml_process_screen(xmlTextReaderPtr reader, 
		struct config* config)
{
	const xmlChar *name, *value;
	struct config_screen *screen = NULL;
	struct config_screen **screens;
	int screen_index = 0;
	int line = xmlutil_get_current_line(reader);
	static int first_screen_entry = 1;

	if ((screen = malloc(sizeof(struct config_screen))) == NULL) {
		rfbLogPerror("config_xml_process_screen: malloc");
		goto error;
	}
	memset(screen, 0, sizeof(struct config_screen));
	screen->vnc_port      = CONFIG_DEFAULT_PORT;
	screen->show_cursor   = CONFIG_DEFAULT_SHOW_CURSOR;
	screen->http_port     = CONFIG_NO_PORT;
	screen->mouse_control = CONFIG_DEFAULT_MOUSE_CONTROL;
	if ((screen->name = config_get_screen_name()) == NULL) {
		goto error;
	} 
			
	while (xmlTextReaderMoveToNextAttribute(reader) > 0) {
		if (xmlutil_read_attribute(reader, line, &name, &value) < 0) {
			goto error;
		}
		if (strcasecmp(name, CONFIG_ATTR_THEME_DIR) == 0) {
			if ((screen->theme_dir = strdup(value)) == NULL) {
				rfbLogPerror("config_xml_process_screen: strdup");
				goto error;
			}
			if (xmlutil_check_dir_path(&(screen->theme_dir)) < 0) {
				goto error;
			}
			rfbLog("Screen %p - theme_dir: %s\n", screen, screen->theme_dir);
		}
		else if (strcasecmp(name, CONFIG_ATTR_VNC_PORT) == 0) {
			screen->vnc_port = atoi(value);
			rfbLog("Screen %p - vnc_port: %d\n", screen,
					screen->vnc_port);
		}
		else if (strcasecmp(name, CONFIG_ATTR_HTTP_PORT) == 0) {
			screen->http_port = atoi(value);
			rfbLog("Screen %p - http_port: %d\n", screen,
					screen->http_port);
		}
		else if (strcasecmp(name, CONFIG_ATTR_SHOW_CURSOR) == 0) {
			screen->show_cursor = atoi(value);
			rfbLog("Screen %p - show_cursor: %d\n", screen,
					screen->show_cursor);
		}
		else if (strcasecmp(name, CONFIG_ATTR_MOUSE_CONTROL) == 0) {
			screen->mouse_control = atoi(value);
			rfbLog("Screen %p - mouse_control: %d\n", screen,
					screen->mouse_control);
		}
		if (strcasecmp(name, CONFIG_ATTR_SCREEN_NAME) == 0) {
			free(screen->name);
			if ((screen->name = strdup(value)) == NULL) {
				rfbLogPerror("config_xml_process_screen: strdup");
				goto error;
			}
			rfbLog("Screen %p - name: %s\n", screen, screen->name);
		}
	}

	if (first_screen_entry) {
		free(config->screens[0]->theme_dir);
		free(config->screens[0]->name);
		free(config->screens[0]);
		config->screens[0] = screen;
		first_screen_entry = 0;
	} 
	else {
		screens = realloc(config->screens, (config->screen_count + 1)
			* sizeof(struct config_screen *));
		if (screens == NULL) {
			rfbLogPerror("config_xml_process_screen: realloc");
			goto error;
		}
		config->screens = screens;
		config->screens[config->screen_count] = screen;
		config->screen_count++;
	} 

	return 0;

error:
	if (screen != NULL) {
		free(screen->theme_dir);
		free(screen->name);
		free(screen);
	}
	return -1;
}

/**
 * Processes a node in a configuration XML document.
 * @param reader  the XML reader
 * @param config  configuration data
 * @return 0 on success or -1 if an error occured
 */
static int config_xml_process_node(xmlTextReaderPtr reader, struct config* config)
{
	const xmlChar *name;
	int ret = 0;

	name  = xmlTextReaderConstName(reader);
	if (strcasecmp(name, CONFIG_ELEM_COMMON) == 0) {
		ret = config_xml_process_common(reader, config);
	}
	else if (strcasecmp(name, CONFIG_ELEM_SERVER) == 0) {
		ret = config_xml_process_screen(reader, config);
	}

	return ret;
}

/**
 * Sets the default configuration values.
 * @param config  configuration data
 * @return 0 on success or -1 if an error occured
 */
static int config_set_defaults(struct config* config)
{
	if (config == NULL) {
		return -1;
	}
	memset(config, 0, sizeof(struct config));
	config->refresh = CONFIG_DEFAULT_REFRESH;
	config->key_repeat_delay = CONFIG_DEFAULT_KEY_DELAY;
	config->screen_count = 1;
	config->screens = malloc(sizeof(struct config_screen *));
	if (config->screens == NULL) {
		rfbLogPerror("config_set_defaults: malloc");
		return -1;
	}
	if ((config->screens[0] = malloc(sizeof(struct config_screen))) == NULL) {
		rfbLogPerror("config_set_defaults: malloc");
		return -1;
	}
	memset(config->screens[0], 0, sizeof(struct config_screen));
	config->screens[0]->vnc_port = CONFIG_DEFAULT_PORT;
	config->screens[0]->show_cursor = CONFIG_DEFAULT_SHOW_CURSOR;
	config->screens[0]->mouse_control = CONFIG_DEFAULT_MOUSE_CONTROL;
	if ((config->screens[0]->name = config_get_screen_name()) == NULL) {
		return -1;
	} 

	return 0;
}

/**
 * Checks correctness of the configuration data.
 * @param config  configuration data
 * @return 0 on success or -1 if an error occured
 */ 
static int config_check(struct config* config) 
{
	int i, j, port1, port2;

	/* Check common values */

	if (config->refresh < 0) {
		rfbLog("ERROR: Display refresh value is negative\n");
		return -1;
	}
	if (config->key_repeat_delay < 0) {
		rfbLog("ERROR: Key repeat delay value is negative\n");
		return -1;
	}

	/* Check vnc listener ports */

	for (i = 0; i < config->screen_count; i++) {
		port1 = config->screens[i]->vnc_port;
		if (port1 < CONFIG_MIN_PORT || port1 > CONFIG_MAX_PORT) {
			rfbLog("ERROR: Screen %p - vnc listener port out of range (%d)\n",
					config->screens[i], port1);
			return -1;
		}
		for  (j = 0; j < config->screen_count; j++) {
			port2 = config->screens[j]->vnc_port;
			if (i != j && port1 == port2) {
				rfbLog("ERROR: Screens %p & %p - identical vnc listener ports\n",
						config->screens[i],  config->screens[j]);
				return -1;
			}
		}
	}

	/* Check http listener ports */

	for (i = 0; i < config->screen_count; i++) {
		port1 = config->screens[i]->http_port;
		if (port1 == CONFIG_NO_PORT) {
			continue;
		}
		if (port1 < CONFIG_MIN_PORT || port1 > CONFIG_MAX_PORT) {
			rfbLog("ERROR: Screen %p - http listener port out of range (%d)\n",
					config->screens[i], port1);
			return -1;
		}
		for  (j = 0; j < config->screen_count; j++) {
			port2 = config->screens[j]->http_port;
			if (i != j && port1 == port2) {
				rfbLog("ERROR: Screens %p & %p - identical http listener ports\n",
						config->screens[i],  config->screens[j]);
				return -1;
			}
		}
	}

	return 0;
}

/**
 * Parses the data of an XML configuration file.
 * @param config_file  configuration file path
 * @return a pointer to a #config structure or NULL if an error occured
 */
struct config *config_load(char *config_file)
{
	xmlTextReaderPtr reader;
	char *config_xml = NULL;
	struct config *config = NULL;
	int ret;

	config = (struct config *)malloc(sizeof(struct config));
	if (config == NULL) {
		rfbLogPerror("config_load: malloc");
		goto error;
	}
	if (config_set_defaults(config) < 0) {
		goto error;
	}
	rfbLog("Default configuration values set\n");

	if (config_file != NULL) {
		config_xml = config_file;
	}
	else {
		config_xml = CONFIG_DEFAULT_FILE;
	}
	reader = xmlReaderForFile(config_xml, NULL, 0);
	if (reader == NULL) {
		rfbLog("Reading of config file '%s' failed\n", config_xml);
		return config;
	}

	rfbLog("Parsing configuration file '%s'\n", config_xml);
	xmlTextReaderSetErrorHandler(reader, xmlutil_error_func, NULL);
	
	while ((ret = xmlTextReaderRead(reader)) == 1) {
		if (config_xml_process_node(reader, config) < 0) {
			goto error;
		}
	}
	if (ret != 0 || config_check(config) < 0) {
		goto error;
	}

	xmlFreeTextReader(reader);
	xmlCleanupParser();
	rfbLog("Configuration file '%s' loaded successfully\n", config_xml);
	
	return config;

error:
	config_cleanup(config);
	if (reader != NULL) {
		xmlFreeTextReader(reader);
		xmlCleanupParser();
	}
	rfbLog("ERROR: Loading of configuration file '%s' failed\n", config_xml);

	return NULL;
}

/**
 * Configuration cleanup. Frees allocated memory.
 * @param config  configuration data
 */
void config_cleanup(struct config *config) 
{
	int i;

	if (config == NULL) {
		return;
	}
	if (config->screens != NULL) {
		for (i = 0; i < config->screen_count; i++) {
			if (config->screens[i] != NULL) {
				free(config->screens[i]->theme_dir);
				free(config->screens[i]->name);
				free(config->screens[i]);
			}
		}
		free(config->screens);
	}
	free(config->auth_file);
	free(config);
}

