/*
 * cover.c - Handles cover art.
 *
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/**
 * \file cover.c
 * Handles cover art. Checks for cover art files in the appropriate directory
 * and reads them.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <rfb/rfb.h>
#include "cover.h"

/**
 * Path of cover art in JPG format.
 */
#define COVER_JPG "/tmp/currcover.jpg"

/**
 * Path of cover art in PNG format.
 */
#define COVER_PNG "/tmp/currcover.png"

/**
 * Checks for cover art in the appropriate directory. If new cover
 * art is avialable it is stored in the given #img_bitmap structure. 
 * @param cover bitmap data (return value)
 * @return possible return values:\n
 *         #COVER_OLD, #COVER_NONE, #COVER_NEW, #COVER_ERR
 */
int cover_check(struct img_bitmap **cover)
{
	struct stat file_stat;
	int cover_status = COVER_OLD;
	char *current_cover = NULL;
	static time_t ts_last_check = 0;
	static time_t ts_last_error = 0;

	if (stat(COVER_JPG, &file_stat) == 0) {
		current_cover = COVER_JPG;
	}
	else if (stat(COVER_PNG, &file_stat) == 0) {
		current_cover = COVER_PNG;
	}
	else {
		return COVER_NONE;
	}

	if (difftime(file_stat.st_mtime, ts_last_check) > 0.0f) {
		if (access(current_cover, R_OK) < 0) {
			rfbLogPerror("cover_check_and_read: access");
			cover_status = COVER_ERR;
		}
		if (difftime(file_stat.st_mtime, ts_last_error) > 0.0f) {
			rfbLog("New cover art available. Loading...\n");
			*cover = img_read(current_cover, IMG_PRESERVE_ALPHA);
			if (*cover != NULL) {
				ts_last_check = file_stat.st_mtime;
				cover_status = COVER_NEW;
			} 
			else {
				ts_last_error = file_stat.st_mtime;
				cover_status = COVER_ERR;
			}
		}
	}

	return cover_status;
}

