/*
 * display.c - Hifidelio display access.
 *
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/**
 * \file display.c
 * Hifidelio display access. Reads the display content into a display buffer,
 * identifies update regions and allows its manipulation with drawing routines.
 */

#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/time.h>
#include <rfb/rfb.h>
#include "display.h"
#include "font.h"
#include "timeutil.h"

/**
 * Display device (framebuffer). Display data is represented with 8 pixels per
 * byte, 50 bytes per row and 160 rows.
 */
#define DISPLAY_DEV_FB "/dev/fb0"

/**
 * Message font corresponding to include file font.h
 */
#define MESSAGE_FONT medium_font

/**
 * Line thickness of message border (pixels).
 */
#define MESSAGE_BORDER_PX 1

/**
 * Distance between message border and text (pixels).
 */
#define MESSAGE_TEXT_EDGE_PX 10

/**
 * Distance between message lines (pixels).
 */
#define MESSAGE_LINE_SPACING_PX 2

/**
 * Display message data.
 */
struct display_message {
	struct timeval msg_ts;    /**< Timestamp of message creation. */
	struct timeval msg_delay; /**< Display duration of message. */
	int line_count;           /**< Line count. */
	int max_line_length;      /**< Maximum line length. */
	char **lines;             /**< Message lines. */
};

/**
 * Currently displayed message.
 */
static struct display_message *msg_pending = NULL;

/**
 * Mutex variable to synchronize access to #msg_pending.
 */
static pthread_mutex_t msg_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * Reads the display content into a given buffer. The content is stored with
 * 8 pixels per byte.
 * @param buffer  buffer to store the display content
 * @param info    framebuffer device information
 * @return 0 on success or -1 if an error occured
 */
static int display_read(unsigned char *buffer, struct fb_var_screeninfo *info)
{
	int fd_fb;
	int bytes = DISPLAY_WIDTH_BYTES * DISPLAY_HEIGHT;

	if ((fd_fb = open(DISPLAY_DEV_FB, O_RDONLY)) < 0) {
		rfbLogPerror("display_read: open");
		goto error;
	}
	if (read(fd_fb, buffer, bytes) < 0) {
		rfbLogPerror("display_read: read");
		goto error;
	}
	if (ioctl(fd_fb, FBIOGET_VSCREENINFO, info)) {
		rfbLogPerror("display_read: ioctl");
		goto error;
	}
	close(fd_fb);
	return 0;

error:
	if (fd_fb > -1) {
		close(fd_fb);
	}
	return -1;
}

/**
 * Fills a rectangle in a display buffer with the given color value.
 * @param x               upper left x-coordinate of rectangle
 * @param y               upper left y-coordinate of rectangle
 * @param width           width of rectangle
 * @param height          height of rectangle
 * @param color           color value; 0 = dark, 1 = light
 * @param display_buffer  display buffer
 * @return 0 on success or -1 if an error occured
 */
static int display_fill_rect(int x, int y, int width, int height,
		int color, unsigned char *display_buffer)
{
	int i, j;
	unsigned char *disp_ptr_off, *disp_ptr;
	unsigned char start_bit;
	unsigned char disp_byte;
	
	if (x < 0 || width <= 0 || x + width > DISPLAY_WIDTH 
			|| y < 0 || height <= 0 || y + height > DISPLAY_HEIGHT) {
		return -1;
	}

	disp_ptr_off = display_buffer + (y * DISPLAY_WIDTH_BYTES) + x / 8;
	start_bit = x % 8;

	if ((start_bit + width) <= 8) {

		/* One byte per row */

		disp_byte = 0;
		for (i = 0; i < width; i++) {
			disp_byte = disp_byte >> 1;
			disp_byte = (0x80 | disp_byte);
		}		
		disp_byte = disp_byte >> start_bit;
		if (color) {
			disp_byte = ~disp_byte;		
		}
		
		for (j = 0; j < height; j++) {
			disp_ptr = disp_ptr_off + j * DISPLAY_WIDTH_BYTES;
			if (!color) {
				*disp_ptr = *disp_ptr |	disp_byte;
			}
			else {
				*disp_ptr = *disp_ptr &	disp_byte;
			}
		}		
	}
	else {
	
		/* Multiple bytes per row */

		for (j = 0; j < height; j++) {
			disp_ptr = disp_ptr_off + j * DISPLAY_WIDTH_BYTES;
			i = 0;
			while (i < width) {
				if (i == 0) {
					disp_byte = 0xff >> start_bit;
					i = 8 - start_bit;
				}
				else if (width - i >= 8) {
					disp_byte = 0xff;
					i += 8;
				}
				else if (width - i < 8) {
					disp_byte = 0xff << (8 - (width - i)); 
					i += width - i;
				}
				if (!color) {
					*disp_ptr = *disp_ptr |	disp_byte;
					disp_ptr++;
				}
				else {
					disp_byte = ~disp_byte;		
					*disp_ptr = *disp_ptr &	disp_byte;
					disp_ptr++;
				}
			}
		}
	}

	return 0;
}

/**
 * Blits a character with a given font and color into a display buffer.
 * @param x               upper left x-coordinate of character
 * @param y               upper left y-coordinate of character
 * @param c               the character
 * @param color           color value; 0 = dark, 1 = light
 * @param font            the font to use
 * @param display_buffer  display buffer
 * @return 0 on success or -1 if an error occured
 */
static int display_blit_char(int x, int y, char c, int color,
		struct font *font, unsigned char *display_buffer)
{
	int i, j;
	int offset_char_bits, offset_disp_bits;
	unsigned char *font_ptr_off, *font_ptr;
	unsigned char *disp_ptr_off, *disp_ptr;
	unsigned char font_bit_init, disp_bit_init;
	unsigned char font_bit, disp_bit;

	if (x < 0 || x + font->char_width > DISPLAY_WIDTH 
			|| y < 0 || y + font->char_height > DISPLAY_HEIGHT) {
		return -1;
	}

	font_ptr_off = font->bytes + (c / font->chars_row) * font->char_width
			* font->char_height * font->chars_row / 8
			+ (c % font->chars_row) * font->char_width / 8;

	disp_ptr_off = display_buffer + (y * DISPLAY_WIDTH_BYTES) + x / 8;
	font_bit_init = (font->char_width * c) % 8;
	disp_bit_init = x % 8;

	for (j = 0; j < font->char_height; j++) {
		font_ptr = font_ptr_off + j * (font->bitmap_width >> 3);
		disp_ptr = disp_ptr_off + j * DISPLAY_WIDTH_BYTES;
		font_bit = font_bit_init;
		disp_bit = disp_bit_init;
		for (i = 0; i < font->char_width; i++) {
			if ((0x80 >> font_bit) & *font_ptr) {
				if (!color) {
					*disp_ptr = (0x80 >> disp_bit) | *disp_ptr;
				}
				else {
					*disp_ptr = (~(0x80 >> disp_bit)) & *disp_ptr;					
				}
			}
			font_bit++;
			if (font_bit == 8) {
				font_bit = 0;
				font_ptr++;
			}
			disp_bit++;
			if (disp_bit == 8) {
				disp_bit = 0;
				disp_ptr++;
			}
		}
	}

	return 0;
}

/**
 * Blits text with a given font and color into a display buffer.
 * @param x               upper left x-coordinate of text
 * @param y               upper left y-coordinate of text
 * @param text            text string
 * @param color           color value; 0 = dark, 1 = light
 * @param font            the font to use
 * @param display_buffer  display buffer
 * @return 0 on success or -1 if an error occured
 */
static int display_blit_text(int x, int y, char* text, int color,
		struct font *font, unsigned char *display_buffer)
{
	int i;

	for (i = 0; i < strlen(text); i++) {
		if (display_blit_char(x + (font->char_width * i), y, text[i], color,
				font, display_buffer) < 0) {
			return -1;
		}
	}

	return 0;
}

/**
 * Draws a message with the given text into a display buffer.
 * @param msg             display message
 * @param display_buffer  display buffer
 * @return 0 on success or -1 if an error occured
 */
static int display_draw_message(struct display_message *msg, 
		unsigned char *display_buffer)
{
	int i, x, y, width, height;
	struct font *msg_font = &MESSAGE_FONT;

	if (msg == NULL || display_buffer == NULL
			|| (msg != NULL && msg->line_count == 0)) {
		goto error;
	}

	/* Draw panel */
	
	height = 2 * MESSAGE_BORDER_PX + 2 * MESSAGE_TEXT_EDGE_PX
			+ msg->line_count * msg_font->char_height
			+ (msg->line_count - 1) * MESSAGE_LINE_SPACING_PX;
	width  = 2 * MESSAGE_BORDER_PX + 2 * MESSAGE_TEXT_EDGE_PX
			+ msg->max_line_length * msg_font->char_width;
	y = (DISPLAY_HEIGHT >> 1) - (height >> 1);
	x = (DISPLAY_WIDTH  >> 1) - (width  >> 1);
	display_fill_rect(x, y, width, height, 1, display_buffer);

	/* Draw border */

	display_fill_rect(
			x, 	y, 
			MESSAGE_BORDER_PX, height, 
			0, display_buffer
		);
	display_fill_rect(
			x, y, 
			width, MESSAGE_BORDER_PX, 
			0, display_buffer
		);
	display_fill_rect(
			x, y + height - MESSAGE_BORDER_PX, 
			width, MESSAGE_BORDER_PX, 
			0, display_buffer
		);
	display_fill_rect(
			x + width - MESSAGE_BORDER_PX, y, 
			MESSAGE_BORDER_PX, height, 
			0, display_buffer
		);

	/* Draw message lines */

	y += MESSAGE_BORDER_PX + MESSAGE_TEXT_EDGE_PX;
	for (i = 0; i < msg->line_count; i++) {
		width = strlen(msg->lines[i]) * msg_font->char_width; 
		display_blit_text(
				(DISPLAY_WIDTH >> 1) - (width >> 1),
				y + i * (msg_font->char_height + MESSAGE_LINE_SPACING_PX),
				msg->lines[i], 0, msg_font, display_buffer
			);
	}

	return 0;

error:
	return -1;
}


/**
 * Message data cleanup. Frees allocated memory.
 * @param msg  message data
 */
static void display_message_cleanup(struct display_message *msg)
{
	int i;

	if (msg == NULL) {
		return;
	}

	if (msg->lines != NULL) {
		for (i = 0; i < msg->line_count; i++) {
			free(msg->lines[i]);
		}
		free(msg->lines);
	}
	free(msg);
}

/**
 * Cuts a message line, if it doesn't fits on the display.
 * @param line  the message line
 * @return formatted message line
 */
static char *display_shorten_message_line (char *line)
{
	int str_len, str_len_px, max_len_px;
	struct font *msg_font = &MESSAGE_FONT;
	char *formatted_line;

	if (line == NULL) {
		goto error;
	}

	str_len_px = msg_font->char_width * strlen(line); 
	max_len_px = DISPLAY_WIDTH - 2 * MESSAGE_BORDER_PX
			- 2 * MESSAGE_TEXT_EDGE_PX;

	if (str_len_px > max_len_px) {
		str_len = max_len_px / msg_font->char_width;
		formatted_line = (char *)malloc(str_len + 1);
		if (formatted_line == NULL) {
			rfbLogPerror("display_shorten_message_line: malloc");
			goto error;
		}
		memset(formatted_line, 0, str_len + 1);
		strncpy(formatted_line, line, str_len - 3);
		strcat(formatted_line, "...");
	}
	else {
		formatted_line = strdup(line);
		if (formatted_line == NULL) {
			rfbLogPerror("display_shorten_message_line: strdup");
			goto error;
		}
	}
	
	return formatted_line;

error:
	return NULL;
}

/**
 * Prepares a message to be shown on the virtual display.
 * @param message   the message
 * @param duration  display duration of message in milliseconds
 * @return 0 on success or -1 if an error occured
 */
int display_show_message(char *message, int duration)
{
	char *token, *tokens = NULL, *delimiters = "\r\n";
	char **lines;
	int max_lines, line_length = 0;
	struct font *msg_font = &MESSAGE_FONT;

	if (message == NULL) {
		goto error;
	}

	pthread_mutex_lock(&msg_mutex);

	if (msg_pending != NULL) {
		display_message_cleanup(msg_pending);
	}

	/* Initialize message structure */

	msg_pending = (struct display_message *)malloc(
			sizeof(struct display_message));
	if (msg_pending == NULL) {
		rfbLogPerror("display_show_message: malloc");
		goto error;
	}
	memset(msg_pending, 0, sizeof(struct display_message));			
	gettimeofday(&(msg_pending->msg_ts), NULL);
	timeutil_get_timeval_from_msec(duration, &(msg_pending->msg_delay));

	/* Extract message lines */

	max_lines = (DISPLAY_HEIGHT + MESSAGE_LINE_SPACING_PX -
			2 * MESSAGE_BORDER_PX - 2 * MESSAGE_TEXT_EDGE_PX) 
			/ (msg_font->char_height + MESSAGE_LINE_SPACING_PX);

	if ((tokens = strdup(message)) == NULL) {
		rfbLogPerror("display_show_message: strdup");
		goto error;
	}	
	token = strtok(tokens, delimiters);
	
	while (token != NULL && msg_pending->line_count < max_lines) {
		lines = realloc(msg_pending->lines,
				(msg_pending->line_count + 1) * sizeof(char *));
		if (lines == NULL) {
			rfbLogPerror("display_show_message: realloc");
			goto error;
		}
		msg_pending->lines = lines;
		msg_pending->lines[msg_pending->line_count] =
				display_shorten_message_line(token);
		if (lines[msg_pending->line_count] == NULL) {
			goto error;
		}
		line_length = strlen(msg_pending->lines[msg_pending->line_count]);
		if (msg_pending->max_line_length < line_length) {
			msg_pending->max_line_length = line_length;
		}
		msg_pending->line_count++;
		token = strtok(NULL, delimiters);
	}

	pthread_mutex_unlock(&msg_mutex);

	free(tokens);
	return 0;

error:
	free(tokens);
	if (msg_pending != NULL) {
		display_message_cleanup(msg_pending);
		msg_pending = NULL;
	}
	return -1;
}

/**
 * Initialize display data structure.
 * @return a display data structure or NULL if an error occured
 */
struct display_data *display_get_data()
{
	unsigned char *display_buffer = NULL;
	struct display_data *data = NULL;

	data = (struct display_data *)malloc(sizeof(struct display_data));
	if (data == NULL) {
		rfbLogPerror("display_get_data: malloc");
		goto error;
	}
	memset(data, 0, sizeof(struct display_data));
	
	display_buffer = (unsigned char *)malloc(DISPLAY_WIDTH_BYTES
			* DISPLAY_HEIGHT);
	if (display_buffer == NULL) {
		rfbLogPerror("display_get_data: malloc");
		goto error;
	}
	data->display_buffer = display_buffer;
	
	return data;

error:
	free(display_buffer);
	free(data);

	return NULL;
}

/**
 * Frees the memory associated with display update regions.
 * @param data  display data
 */
static void display_update_regions_cleanup(struct display_data *data) 
{
	int i;

	if (data == NULL) {
		return;
	}

	if (data->update_regions != NULL) {
		for (i = 0; i < data->update_regions_count; i++) {
			free(data->update_regions[i]);
		}
		free(data->update_regions);
		data->update_regions = NULL;
	}
	data->update_regions_count = 0;
}

/**
 * Reads the display content and identifies the update regions.
 * @param data  display data
 * @return 0 on success or -1 if an error occured
 */
int display_update(struct display_data *data)
{
	int x, y;
	int x1 = -1, x2, y1, y2, line_changed;
	static unsigned char display[DISPLAY_WIDTH_BYTES * DISPLAY_HEIGHT];
	unsigned char *dp_ptr  = NULL;	
	unsigned char *dpb_ptr = NULL;
	struct display_region *update_region   = NULL;
	struct display_region **update_regions = NULL;

	if (data == NULL) {
		rfbLog("ERROR: No valid display data structure\n");
		goto error;
	}
	dpb_ptr = data->display_buffer;
	display_update_regions_cleanup(data);

	if (display_read(display, &(data->info)) < 0) {
		goto error;
	}

	pthread_mutex_lock(&msg_mutex);

	if (msg_pending != NULL) {
		if (timeutil_interval_exceeded(&(msg_pending->msg_ts), 
				&(msg_pending->msg_delay))) {
			display_message_cleanup(msg_pending);
			msg_pending = NULL;
		}
		else {
			display_draw_message(msg_pending, display);
		}
	}

	pthread_mutex_unlock(&msg_mutex);
	
	dp_ptr = display;

	for (y = 0; y < DISPLAY_HEIGHT; y++) {
		line_changed = 0;
		for (x = 0; x < DISPLAY_WIDTH_BYTES; x++) {
			if (*dp_ptr != *dpb_ptr) {
				if (x1 == -1) {
					x1 = x; y1 = y;	x2 = x;	y2 = y;
				}
				if (x < x1) {
					x1 = x;
				}
				if (x > x2) {
					x2 = x;
				}
				if (y > y2) {
					y2 = y;
				}
				*dpb_ptr = *dp_ptr;
				line_changed = 1;
			}
			dp_ptr++;
			dpb_ptr++;
		}
		if ((!line_changed || y == DISPLAY_HEIGHT - 1) && x1 != -1) {		
			update_regions = realloc(data->update_regions,
					(data->update_regions_count + 1)
					* sizeof(struct display_region *));		
			if (update_regions == NULL) {
				rfbLogPerror("display_update: realloc");
				goto error;
			}
			data->update_regions = update_regions;
			update_region = (struct display_region *)malloc(
					sizeof(struct display_region));
			if (update_region == NULL) {
				rfbLogPerror("display_update: malloc");
				goto error;
			}
			update_region->y1 = y1;
			update_region->x1 = x1 << 3;
			update_region->y2 = y2;
			update_region->x2 = (x2 + 1) << 3;
			if (update_region->x2 == DISPLAY_WIDTH) {
				update_region->x2 = update_region->x2 - 1;
			}
			data->update_regions[data->update_regions_count] = update_region;
			data->update_regions_count++;			
			x1 = -1;
		}
	}
	return 0;

error:
	return -1;
}

/**
 * Display data cleanup. Frees allocated memory.
 * @param data  display data
 */
void display_cleanup(struct display_data *data) 
{
	int i;

	if (data == NULL) {
		return;
	}

	display_update_regions_cleanup(data);
	free(data->display_buffer);
	free(data);
}

