/*
 * img.c - Reads image files (BMP, PNG, JPG).
 * 
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/**
 * \file img.c
 * Reads image files (BMP, PNG, JPG).
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <endian.h>
#include <rfb/rfb.h>
#include "img.h"
#include "png.h"
#include "jpeglib.h"

/** 
 * \def e16(A)
 * Endian conversion for 16-bit values.
 */

/** 
 * \def e32(A)
 * Endian conversion for 32-bit values.
 */

#if defined __BIG_ENDIAN
#define e16(A) ( \
		(((unsigned short)(A) & 0xff00) >> 8) | \
		(((unsigned short)(A) & 0x00ff) << 8)   \
		)
#define e32(A) ( \
		(((unsigned long)(A) & 0xff000000) >> 24) | \
		(((unsigned long)(A) & 0x00ff0000) >> 8 ) | \
		(((unsigned long)(A) & 0x0000ff00) << 8 ) | \
		(((unsigned long)(A) & 0x000000ff) << 24)   \
		)
#else
#define e16(A) A
#define e32(A) A
#endif

/**
 * BMP file extension. 
 */
#define IMG_BMP_EXT ".bmp"

/**
 * BMP identifier (first 2 bytes of image file).
 */
#define IMG_BMP_IDENTIFIER  0x4d42 /* "BM" */

/**
 * BMP header data.
 */
struct bmp_header {
	unsigned short type;             /**< Magic identifier. */
	unsigned long  file_size;        /**< File size in bytes. */
	unsigned short reserved_1;
	unsigned short reserved_2;
	unsigned long  offset;           /**< Offset to image data, bytes. */
	unsigned long  header_size;      /**< Header size in bytes. */
	unsigned long  width;            /**< Width of image. */
	unsigned long  height;           /**< Height of image. */
	unsigned short planes;           /**< Number of color planes. */   
	unsigned short bits;             /**< Bits per pixel. */
	unsigned long  compression;      /**< Compression type. */
	unsigned long  imagesize;        /**< Image size in bytes. */
	unsigned long  x_resolution;     /**< X pixels per meter. */
	unsigned long  y_resolution;     /**< Y pixels per meter. */
	unsigned long  n_colors;         /**< Number of colors. */
	unsigned long  important_colors; /**< Important colors. */
} __attribute__ ((packed));

/**
 * PNG file extension. 
 */
#define IMG_PNG_EXT ".png"

/**
 * Number of check bytes for PNG images. 
 */
#define IMG_PNG_CHECK_BYTES 4

/**
 * JPG file extension. 
 */
#define IMG_JPG_EXT ".jpg"

/**
 * JPG error handling data.
 */
struct img_jpg_error_mgr {
	struct  jpeg_error_mgr pub; /**< "public" fields. */
	jmp_buf setjmp_buffer;      /**< For return to caller. */
};

/**
 * JPG error handling data.
 */
typedef struct img_jpg_error_mgr *img_jpg_error_ptr;

/**
 * Reads a 24bpp BMP file.
 * @param filename      BMP filename
 * @param bitmap        bitmap data (return value)
 * @return 0 on success or -1 if an error occured
 */
static int img_read_bmp(char *filename, struct img_bitmap *bitmap)
{
	struct bmp_header header;
	FILE *fp;
	long x, y, width, height, offset;
	long pixels;
	unsigned char *buffer  = NULL;
	unsigned char *buf_ptr = NULL;
	unsigned char *bmp_ptr = NULL;

	/* Read header and check bitmap format */

	if ((fp = fopen(filename, "r")) == NULL) {
		rfbLogPerror("img_read_bmp: fopen");
		goto error;
	}
	if (fread(&header, sizeof(struct bmp_header), 1, fp) != 1) {
		goto error;
	}
	if (e16(header.type) != IMG_BMP_IDENTIFIER || e16(header.bits) != 24) {
		rfbLog("ERROR: BMP file '%s' with wrong format (non-24bpp)\n", 
				filename);
		goto error;
	}

	/* Allocate bitmap memory */

	width  = e32(header.width);
	height = e32(header.height);
	pixels = width * height * 3;

	buffer = malloc(pixels);
	if (buffer == NULL) {
		rfbLogPerror("img_read_bmp: malloc");
		goto error;
	}
	bitmap->pixels = bmp_ptr = malloc(width * height * 4);
	if (bitmap->pixels == NULL) {
		rfbLogPerror("img_read_bmp: malloc");
		goto error;
	}

	/* Read bitmap data */

	if (fseek(fp, e32(header.offset), SEEK_SET) <  0) {
		rfbLogPerror("img_read_bmp: fseek");
		goto error;
	}
	if (fread(buffer, pixels, 1, fp) != 1) {
		goto error;
	}
	buf_ptr = buffer;

	/* Store bitmap data into img_bitmap structure */

	for (y = height - 1; y >= 0; y--) {
		for (x = 0; x < width; x++) {
			offset = ((y * width) << 2) + (x << 2);
			*(bmp_ptr + offset + 3) = 0xff;
			*(bmp_ptr + offset + 2) = *(buf_ptr++); 
			*(bmp_ptr + offset + 1) = *(buf_ptr++); 
			*(bmp_ptr + offset)     = *(buf_ptr++); 
		}
	}
	bitmap->width  = width;
	bitmap->height = height;
	bitmap->alpha_channel = 0;
	free(buffer);
	fclose(fp);

	rfbLog("Bitmap file: '%s'\n", filename);
	rfbLog("Bitmap resolution: %d x %d\n", width, height);

	return 0;

error:
	if (fp != NULL) {
		fclose(fp);
	}
	free(bitmap->pixels);
	free(buffer);

	rfbLog("ERROR: Reading of BMP file '%s' failed\n", filename);

	return -1;
}

/**
 * PNG error function.
 * @param png_ptr    pointer to PNG context
 * @param error_msg  error message
 */
static void img_png_error(png_structp png_ptr, png_const_charp error_msg) 
{
	rfbLog("ERROR: %s\n", error_msg);
}

/**
 * PNG warning function.
 * @param png_ptr      pointer to PNG context
 * @param warning_msg  warning message
 */
static void img_png_warning(png_structp png_ptr, png_const_charp warning_msg)
{
	rfbLog("WARNING: %s\n", warning_msg);
}

/**
 * Reads a 24bpp or 32bpp PNG file.
 * @param filename    PNG filename
 * @param bitmap      bitmap data (return value)
 * @param alpha_mode  preserve or ignore alpha channel
 * @return 0 on success or -1 if an error occured
 */
static int img_read_png(char *filename, struct img_bitmap *bitmap,
		int alpha_mode)
{
	FILE *fp;
	png_structp png_ptr = NULL;
	png_infop info_ptr  = NULL;
	char buf[IMG_PNG_CHECK_BYTES];
	int png_transforms = PNG_TRANSFORM_IDENTITY;
	png_uint_32 width, height;
	int bit_depth, channels;
	png_bytep *row_pointers;
	unsigned char *bmp_ptr = NULL;
	int x, y;

	/* Open file and check PNG signature */

	if ((fp = fopen(filename, "rb")) == NULL) {
		rfbLogPerror("img_read_png: fopen");
		goto error;
	}
	if (fsync(fileno(fp)) < 0) {
		rfbLogPerror("img_read_png: fsync");
	}
	if (fread(buf, 1, IMG_PNG_CHECK_BYTES, fp) != IMG_PNG_CHECK_BYTES) {
		rfbLog("ERROR: Unable to read PNG signature\n");
		goto error;
	}
	if(png_sig_cmp(buf, (png_size_t)0, IMG_PNG_CHECK_BYTES)) {
		rfbLog("ERROR: PNG file with wrong signature\n");
		goto error;
	}

	/* Initialize read and info structures */

	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, 
			img_png_error, img_png_warning);
	if (png_ptr == NULL) {
		rfbLog("ERROR: Unable to create PNG read structure\n");
		goto error;
	}
	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL) {
		rfbLog("ERROR: Unable to create PNG info structure\n");
		goto error;
	}

	/* Set error handling */

	if (setjmp(png_jmpbuf(png_ptr))) {
		goto error;
	}

	/* Read image data */

	png_init_io(png_ptr, fp);
	png_set_sig_bytes(png_ptr, IMG_PNG_CHECK_BYTES);
	png_read_png(png_ptr, info_ptr, png_transforms, png_voidp_NULL);

	channels  = png_get_channels (png_ptr, info_ptr);
	bit_depth = png_get_bit_depth(png_ptr, info_ptr);
	width  = png_get_image_width (png_ptr, info_ptr);
	height = png_get_image_height(png_ptr, info_ptr);	

	/* Store bitmap data into img_bitmap structure */

	if (bit_depth != 8 && (channels != 3 || channels != 4)) {
		rfbLog("ERROR: PNG file '%s' with wrong format (non-24bpp, non-32bpp)\n", 
				filename);
		goto error;
	}
	bitmap->pixels = bmp_ptr = malloc(width * height * 4);
	if (bitmap->pixels == NULL) {
		rfbLogPerror("img_read_png: malloc");
		goto error;
	}
	row_pointers = png_get_rows(png_ptr, info_ptr);
	for (y = 0; y < height; y++) {
		for (x = 0; x < width * channels; x += channels) {
			*(bmp_ptr++) = *(row_pointers[y] + x);
			*(bmp_ptr++) = *(row_pointers[y] + x + 1);
			*(bmp_ptr++) = *(row_pointers[y] + x + 2);
			if (channels == 4 && alpha_mode == IMG_PRESERVE_ALPHA) {
				*(bmp_ptr++) = *(row_pointers[y] + x + 3);
			}
			else {
				*(bmp_ptr++) = 0xff;
			}
		}
	}
	bitmap->width  = width;
	bitmap->height = height;
	if (channels == 4) {
		bitmap->alpha_channel = 1;
	}
	else {
		bitmap->alpha_channel = 0;
	}

	rfbLog("Bitmap file: '%s'\n", filename);
	rfbLog("Bitmap resolution: %d x %d alpha channel: %s \n", width, height,
			(bitmap->alpha_channel ? "yes" : "no"));

	png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
	fclose(fp);

	return 0;

error:
	if (png_ptr != NULL && info_ptr != NULL) {
		png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
	}
	else if(png_ptr != NULL) {
		png_destroy_read_struct(&png_ptr, png_infopp_NULL, png_infopp_NULL);
	}
	if (fp != NULL) {
		fclose(fp);
	}
	free(bitmap->pixels);

	rfbLog("ERROR: Reading of PNG file '%s' failed\n", filename);

	return -1;
}

/**
 * JPG error function.
 * @param cinfo  jpeg decompression parameters
 */ 
static void img_jpg_error(j_common_ptr cinfo)
{
	char buffer[JMSG_LENGTH_MAX];

	img_jpg_error_ptr error = (img_jpg_error_ptr)cinfo->err;
	(*cinfo->err->format_message) (cinfo, buffer);
	rfbLog("ERROR: %s\n", buffer);
	
	longjmp(error->setjmp_buffer, 1);
}

/**
 * Reads a 24bpp JPG file.
 * @param filename  JPG filename
 * @param bitmap    bitmap data (return value)
 * @return 0 on success or -1 if an error occured
 */
static int img_read_jpg(char *filename, struct img_bitmap *bitmap)
{
	FILE *fp;
	struct jpeg_decompress_struct cinfo;
	struct img_jpg_error_mgr jerr;
	JSAMPARRAY buffer;
	int row_stride, i;
	unsigned char *bmp_ptr = NULL;

	/* Open file and set error handling */

	if ((fp = fopen(filename, "rb")) == NULL) {
		rfbLogPerror("img_read_jpg: fopen");
		goto error;
	}
	if (fsync(fileno(fp)) < 0) {
		rfbLogPerror("img_read_jpg: fsync");
	}
	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = img_jpg_error;
	if (setjmp(jerr.setjmp_buffer)) {
		goto error;
	}
	
	/* Initialize the JPEG decompression object and check header. */
	
	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, fp);
	jpeg_read_header(&cinfo, TRUE);

	/* Start decompression */
	
	jpeg_start_decompress(&cinfo);
	if (cinfo.output_components != 3) {
		rfbLog("ERROR: JPG file '%s' with wrong format (non-24bpp)\n",
				filename);
		goto error;
	}
	row_stride = cinfo.output_width * cinfo.output_components;
	buffer = (*cinfo.mem->alloc_sarray) ((j_common_ptr) &cinfo, JPOOL_IMAGE, 
			row_stride, 1);

	/* Store bitmap data into img_bitmap structure */

	bitmap->width  = cinfo.output_width;
	bitmap->height = cinfo.output_height;
	bitmap->pixels = bmp_ptr = malloc(bitmap->width * bitmap->height * 4);
	bitmap->alpha_channel = 0;
	if (bitmap->pixels == NULL) {
		rfbLogPerror("img_read_jpg: malloc");
		goto error;
	}
	while (cinfo.output_scanline < cinfo.output_height) {
		jpeg_read_scanlines(&cinfo, buffer, 1);
		for (i = 0; i < row_stride; i += 3) {
			*(bmp_ptr++) = *(buffer[0] + i);
			*(bmp_ptr++) = *(buffer[0] + i + 1);
			*(bmp_ptr++) = *(buffer[0] + i + 2);
			*(bmp_ptr++) = 0xff;
		}
	}

	/* Finish decompression */

	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);
	fclose(fp);

	rfbLog("Bitmap file: '%s'\n", filename);
	rfbLog("Bitmap resolution: %d x %d\n", bitmap->width, bitmap->height);

	return 0;

error:
	jpeg_destroy_decompress(&cinfo);
	if (fp != NULL) {
		fclose(fp);
	}
	free(bitmap->pixels);

	rfbLog("ERROR: Reading of JPG file '%s' failed\n", filename);

	return -1;
}

/**
 * Reads an image file (BMP, PNG, JPG).
 * @param filename image filename
 * @param alpha_mode preserve or ignore alpha channel
 * @return a pointer to a bitmap or NULL if an error occured
 */
struct img_bitmap *img_read(char *filename, int alpha_mode)
{
	char *ext = NULL;
	struct img_bitmap *bitmap = NULL;
	int read_result = -1;

	if ((ext = strrchr(filename, '.')) == NULL) {
		rfbLog("ERROR: Bitmap file '%s' has no extension\n", filename);
		goto error;
	}

	bitmap = malloc(sizeof(struct img_bitmap));
	memset(bitmap, 0, sizeof(struct img_bitmap));

	if (!strcasecmp(ext, IMG_PNG_EXT)) {
		read_result = img_read_png(filename, bitmap, alpha_mode);
	}
	else if (!strcasecmp(ext, IMG_JPG_EXT)) {
		read_result = img_read_jpg(filename, bitmap);
	}
	else if (!strcasecmp(ext, IMG_BMP_EXT)) {
		read_result = img_read_bmp(filename, bitmap);
	}
	else {
		rfbLog("ERROR: Bitmap file '%s' has an unknown extension\n", 
				filename);
		goto error;
	}
	
	if (read_result == 0) {
		return bitmap;
	}

error:
	img_free(bitmap);
	return NULL;
}

/**
 * Bitmap data cleanup. Frees allocated memory.
 * @param bitmap   bitmap data
 */
void img_free(struct img_bitmap *bitmap)
{
	if (bitmap == NULL) {
		return;
	}
	free(bitmap->pixels);
	free(bitmap);
}
