/*
 * keys.c - Keycode mapping VNC -> Hifidelio.
 *
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/**
 * \file keys.c
 * Keycode mapping VNC -> Hifidelio. Translates X keyboard keycodes into
 * Hifidelio keycodes with modifier.
 */

#include <stdlib.h>
#include "keys.h"

/**
 * Keycode mapping entry.
 */
struct keys_vnc_hf_mapping {
	rfbKeySym vnc_key;      /**< VNC keycode */
	rfbKeySym hf_key;       /**< HF keycode */
	rfbKeySym hf_modifier;  /**< HF key modifier */
};

/**
 * Keycode mapping array.
 */
static struct keys_vnc_hf_mapping key_mapping[] = {
	{ XK_Escape,         SDLK_ESCAPE,       KMOD_NONE  },
	{ XK_F1,             SDLK_F1,           KMOD_NONE  },
	{ XK_F2,             SDLK_F2,           KMOD_NONE  },
	{ XK_F3,             SDLK_F3,           KMOD_NONE  },
	{ XK_F4,             SDLK_F4,           KMOD_NONE  },
	{ XK_F5,             SDLK_F5,           KMOD_NONE  },
	{ XK_F6,             SDLK_F6,           KMOD_NONE  },
	{ XK_F7,             SDLK_F7,           KMOD_NONE  },
	{ XK_F8,             SDLK_F8,           KMOD_NONE  },
	{ XK_F9,             SDLK_F9,           KMOD_NONE  },
	{ XK_F10,            SDLK_F10,          KMOD_NONE  },
	{ XK_F11,            SDLK_F11,          KMOD_NONE  },
	{ XK_F12,            SDLK_F12,          KMOD_NONE  },
	{ XK_Delete,         SDLK_DELETE,       KMOD_NONE  },
	{ XK_1,              SDLK_1,            KMOD_NONE  },
	{ XK_2,              SDLK_2,            KMOD_NONE  },
	{ XK_3,              SDLK_3,            KMOD_NONE  },
	{ XK_4,              SDLK_4,            KMOD_NONE  },
	{ XK_5,              SDLK_5,            KMOD_NONE  },
	{ XK_6,              SDLK_6,            KMOD_NONE  },
	{ XK_7,              SDLK_7,            KMOD_NONE  },
	{ XK_8,              SDLK_8,            KMOD_NONE  },
	{ XK_9,              SDLK_9,            KMOD_NONE  },
	{ XK_0,              SDLK_0,            KMOD_NONE  },
	{ XK_ssharp,         SDLK_WORLD_1,      KMOD_NONE  },
	{ XK_BackSpace,      SDLK_BACKSPACE,    KMOD_NONE  },
	{ XK_Home,           SDLK_HOME,         KMOD_NONE  },
	{ XK_exclam,         SDLK_1,            KMOD_SHIFT },
	{ XK_quotedbl,       SDLK_2,            KMOD_SHIFT },
	{ XK_section,        SDLK_3,            KMOD_SHIFT },
	{ XK_dollar,         SDLK_4,            KMOD_SHIFT },
	{ XK_percent,        SDLK_5,            KMOD_SHIFT },
	{ XK_ampersand,      SDLK_6,            KMOD_SHIFT },
	{ XK_slash,          SDLK_7,            KMOD_SHIFT },
	{ XK_parenleft,      SDLK_8,            KMOD_SHIFT },
	{ XK_parenright,     SDLK_9,            KMOD_SHIFT },
	{ XK_equal,          SDLK_0,            KMOD_SHIFT },
	{ XK_question,       SDLK_WORLD_1,      KMOD_SHIFT },
	{ XK_braceleft,      SDLK_7,            KMOD_ALT   },
	{ XK_bracketleft,    SDLK_8,            KMOD_ALT   },
	{ XK_bracketright,   SDLK_9,            KMOD_ALT   },
	{ XK_braceright,     SDLK_0,            KMOD_ALT   },
	{ XK_backslash,      SDLK_WORLD_1,      KMOD_ALT   },
	{ XK_Tab,            SDLK_PAGEDOWN,     KMOD_NONE  },
	{ XK_q,              SDLK_q,            KMOD_NONE  },
	{ XK_w,              SDLK_w,            KMOD_NONE  },
	{ XK_e,              SDLK_e,            KMOD_NONE  },
	{ XK_r,              SDLK_r,            KMOD_NONE  },
	{ XK_t,              SDLK_t,            KMOD_NONE  },
	{ XK_z,              SDLK_z,            KMOD_NONE  },
	{ XK_u,              SDLK_u,            KMOD_NONE  },
	{ XK_i,              SDLK_i,            KMOD_NONE  },
	{ XK_o,              SDLK_o,            KMOD_NONE  },
	{ XK_p,              SDLK_p,            KMOD_NONE  },
	{ XK_udiaeresis,     SDLK_WORLD_2,      KMOD_NONE  },
	{ XK_plus,           SDLK_PLUS,         KMOD_NONE  },
	{ XK_numbersign,     SDLK_HASH,         KMOD_NONE  },
	{ XK_Page_Up,        SDLK_PAGEUP,       KMOD_NONE  },
	{ XK_Q,              SDLK_q,            KMOD_SHIFT },
	{ XK_W,              SDLK_w,            KMOD_SHIFT },
	{ XK_E,              SDLK_e,            KMOD_SHIFT },
	{ XK_R,              SDLK_r,            KMOD_SHIFT },
	{ XK_T,              SDLK_t,            KMOD_SHIFT },
	{ XK_Z,              SDLK_z,            KMOD_SHIFT },	
	{ XK_U,              SDLK_u,            KMOD_SHIFT },
	{ XK_I,              SDLK_i,            KMOD_SHIFT },
	{ XK_O,              SDLK_o,            KMOD_SHIFT },
	{ XK_P,              SDLK_p,            KMOD_SHIFT },
	{ XK_Udiaeresis,     SDLK_WORLD_2,      KMOD_SHIFT },
	{ XK_asterisk,       SDLK_PLUS,         KMOD_SHIFT },
	{ XK_apostrophe,     SDLK_HASH,         KMOD_SHIFT },
	{ XK_at,             SDLK_q,            KMOD_ALT   },
	{ XK_EuroSign,       SDLK_e,            KMOD_ALT   },
	{ XK_asciitilde,     SDLK_PLUS,         KMOD_ALT   },
	{ XK_Caps_Lock,      SDLK_CAPSLOCK,     KMOD_NONE  },
	{ XK_a,              SDLK_a,            KMOD_NONE  },
	{ XK_s,              SDLK_s,            KMOD_NONE  },
	{ XK_d,              SDLK_d,            KMOD_NONE  },
	{ XK_f,              SDLK_f,            KMOD_NONE  },
	{ XK_g,              SDLK_g,            KMOD_NONE  },
	{ XK_h,              SDLK_h,            KMOD_NONE  },
	{ XK_j,              SDLK_j,            KMOD_NONE  },
	{ XK_k,              SDLK_k,            KMOD_NONE  },
	{ XK_l,              SDLK_l,            KMOD_NONE  },
	{ XK_odiaeresis,     SDLK_WORLD_4,      KMOD_NONE  },
	{ XK_adiaeresis,     SDLK_WORLD_3,      KMOD_NONE  },
	{ XK_Return,         SDLK_RETURN,       KMOD_NONE  },
	{ XK_Page_Down,      SDLK_PAGEDOWN,     KMOD_NONE  },
	{ XK_A,              SDLK_a,            KMOD_SHIFT },
	{ XK_S,              SDLK_s,            KMOD_SHIFT },
	{ XK_D,              SDLK_d,            KMOD_SHIFT },
	{ XK_F,              SDLK_f,            KMOD_SHIFT },
	{ XK_G,              SDLK_g,            KMOD_SHIFT },
	{ XK_H,              SDLK_h,            KMOD_SHIFT },
	{ XK_J,              SDLK_j,            KMOD_SHIFT },
	{ XK_K,              SDLK_k,            KMOD_SHIFT },
	{ XK_L,              SDLK_l,            KMOD_SHIFT },
	{ XK_Odiaeresis,     SDLK_WORLD_4,      KMOD_SHIFT },
	{ XK_Adiaeresis,     SDLK_WORLD_3,      KMOD_SHIFT },
	{ XK_y,              SDLK_y,            KMOD_NONE  },
	{ XK_x,              SDLK_x,            KMOD_NONE  },
	{ XK_c,              SDLK_c,            KMOD_NONE  },
	{ XK_v,              SDLK_v,            KMOD_NONE  },
	{ XK_b,              SDLK_b,            KMOD_NONE  },
	{ XK_n,              SDLK_n,            KMOD_NONE  },
	{ XK_m,              SDLK_m,            KMOD_NONE  },
	{ XK_comma,          SDLK_COMMA,        KMOD_NONE  },
	{ XK_period,         SDLK_PERIOD,       KMOD_NONE  },
	{ XK_minus,          SDLK_MINUS,        KMOD_NONE  },
	{ XK_Up,             SDLK_UP,           KMOD_NONE  },
	{ XK_End,            SDLK_END,          KMOD_NONE  },
	{ XK_Y,              SDLK_y,            KMOD_SHIFT },
	{ XK_X,              SDLK_x,            KMOD_SHIFT },
	{ XK_C,              SDLK_c,            KMOD_SHIFT },
	{ XK_V,              SDLK_v,            KMOD_SHIFT },
	{ XK_B,              SDLK_b,            KMOD_SHIFT },
	{ XK_N,              SDLK_n,            KMOD_SHIFT },
	{ XK_M,              SDLK_m,            KMOD_SHIFT },
	{ XK_semicolon,      SDLK_COMMA,        KMOD_SHIFT },
	{ XK_colon,          SDLK_PERIOD,       KMOD_SHIFT },
	{ XK_underscore,     SDLK_MINUS,        KMOD_SHIFT },
	{ XK_space,          SDLK_SPACE,        KMOD_NONE  },
	{ XK_less,           SDLK_LESS,         KMOD_NONE  },
	{ XK_Left,           SDLK_LEFT,         KMOD_NONE  },
	{ XK_Down,           SDLK_DOWN,         KMOD_NONE  },
	{ XK_Right,          SDLK_RIGHT,        KMOD_NONE  },
	{ XK_greater,        SDLK_LESS,         KMOD_SHIFT },
	{ XK_bar,            SDLK_LESS,         KMOD_ALT   },
	{ XK_KP_Add,         SDLK_KP_PLUS,      KMOD_NONE  },
	{ XK_KP_Subtract,    SDLK_KP_MINUS,     KMOD_NONE  }
};

/**
 * Size of keycode mapping array.
 */
const int keys_mapping_size =
		sizeof(key_mapping)/sizeof(struct keys_vnc_hf_mapping);

/**
 * Comparison function for keycode mapping entries.
 * @param a  mapping entry
 * @param b  mapping entry
 * @return negative if the first argument is less than the second, zero if
 * they are equal, and positive if the first argument is greater
 */
static int keys_compare_mapping(const void *a, const void *b) 
{
	struct keys_vnc_hf_mapping *ma = (struct keys_vnc_hf_mapping *)a;
	struct keys_vnc_hf_mapping *mb = (struct keys_vnc_hf_mapping *)b;
	
	return (ma->vnc_key > mb->vnc_key) - (ma->vnc_key < mb->vnc_key);
}

/**
 * Translates VNC keycodes into Hifidelio keycodes.
 * @param vnc_key      VNC keycode
 * @param hf_key       Hifidelio keycode (return value)
 * @param hf_modifier  Hifidelio key modifier code (return value)
 * @return 0 on success or -1 if an error occured
 */
int keys_translate_vnc_hf(rfbKeySym vnc_key, rfbKeySym *hf_key, 
		rfbKeySym *hf_modifier) 
{
	static int mapping_sorted = 0;
	struct keys_vnc_hf_mapping mapping;
	struct keys_vnc_hf_mapping *result;

	if (!mapping_sorted) {
		qsort(key_mapping, keys_mapping_size, 
				sizeof(struct keys_vnc_hf_mapping), keys_compare_mapping);
		mapping_sorted = 1;
	}

	mapping.vnc_key = vnc_key;
	result = bsearch(&mapping, key_mapping, keys_mapping_size,
			sizeof(struct keys_vnc_hf_mapping), keys_compare_mapping);
	if (result != NULL) {
		*hf_key = result->hf_key;
		*hf_modifier = result->hf_modifier;
	}

	return (result != NULL) - 1;
} 

/**
 * Checks if a given XK keycode belongs to a key modifier.
 * @param xk_key  XK keycode
 * @return 1 if key is a modifier key, else 0
 */
int keys_is_xk_modifier(rfbKeySym xk_key)
{
	return (   xk_key == XK_Shift_L
			|| xk_key == XK_Shift_R
			|| xk_key == XK_Control_L
			|| xk_key == XK_Control_R
			|| xk_key == XK_Meta_L
			|| xk_key == XK_Meta_R
			|| xk_key == XK_Alt_L
			|| xk_key == XK_Alt_R
			|| xk_key == XK_Super_L
			|| xk_key == XK_Super_R
			|| xk_key == XK_Hyper_L
			|| xk_key == XK_Hyper_R  );
}
