/*
 * scale.c - Bitmap scaling functions.
 *
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/**
 * \file scale.c
 * Bitmap scaling functions.
 */

#include <stdio.h>
#include <stdlib.h>
#include <rfb/rfb.h>
#include "scale.h"

/**
 * Calculates scaling vectors that can be used for 'nearest neighbour' image
 * sampling. When magnifying, pixels from the source image are replicated,
 * when minifying (scaling down), pixels from the source image are dropped.
 *
 * \code
 * Example: scale factor = 2.0
 *
 * source
 * 00 01 02 03 04 05 06 07 08 09 10 11 12 13 14 15 16 17 18 19 20 21 ...
 *
 * line
 * 00 00 01 01 02 02 03 03 04 04 05 05 06 06 07 07 08 08 09 09 10 10 ...
 *
 * l_off
 * 00 02 04 06 08 10 12 14 16 18 20 22 24 26 28 30 32 34 36 38 40 42 ...
 *
 * r_off
 * 01 03 05 07 09 11 13 15 17 19 21 23 25 27 29 31 33 35 37 39 41 43 ...
 * \endcode
 *
 * @param orig    original size
 * @param scaled  scaled size
 * @param line    scale vector (return value)
 * @param l_off   scale vector offsets, left to right (return value, optional)
 * @param r_off   scale vector offsets, right to left (return value, optional)
 * @return 0 on success or -1 if an error occured
 */
int scale_vectors(int orig, int scaled, int **line, int **l_off, int **r_off)
{
	int i, a, dx, dy;
	int co = orig;
	int cs = scaled;
	int *v  = NULL;
	int *vl = NULL;
	int *vr = NULL;

	v = (int*)malloc(scaled * sizeof(int));
	if (v == NULL) {
		rfbLogPerror("scale_vectors: malloc");
		goto error;
	}

	if (l_off != NULL) {
		vl = (int*)malloc(orig * sizeof(int));
		if (vl == NULL) {
			rfbLogPerror("scale_vectors: malloc");
			goto error;
		}
	}

	if (r_off != NULL) {
		vr = (int*)malloc(orig * sizeof(int));
		if (vr == NULL) {
			rfbLogPerror("scale_vectors: malloc");
			goto error;
		}
	}

	if (co > cs) {
		dy = 2 * cs;
		dx = a = 2 * co - dy;
		do {
			if (a <= 0) {
				cs--;
				v[cs] = co - 1;
				a += dx;
			} 
			else {
				a -= dy;
			}
		} while (co--);
	}
	else {
		dy = 2 * co;
		dx = a = 2 * cs - dy;
		do {
			cs--;
			v[cs] = co - 1;
			if (a <= 0) {
				co--;
				a += dx;
			} 
			else {
				a -= dy;
			}
		} while (cs);
	}

	*line = v;

	if (vl != NULL) {
		co = cs = 0;
		while (co < orig) {
			if (v[cs] >= co || cs == scaled - 1) {
				vl[co++] = cs;
			}
			else if (cs < scaled - 1) {
				cs++;
			}
		}
		*l_off = vl;
	}

	if (vr != NULL) {
		co = orig - 1;
		cs = scaled - 1;
		while (co >= 0) {
			if (v[cs] <= co || cs == 0) {
				vr[co--] = cs;
			}
			else if (cs > 0) {
				cs--;
			}
		}
		*r_off = vr;
	}

	return 0;

error:
	free(v);
	free(vl);
	free(vr);
	return -1;
}

