/*
 * timeutil.c - Utility functions for time calculations.
 *
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/**
 * \file timeutil.c
 * Utility functions for time calculations.
 */

#include <stdio.h>
#include "timeutil.h"

/**
 * Fills a timeval structure for the given milliseconds value.
 * @param msec  milliseconds
 * @param tv    timeval structure
 * @return 0 on success or -1 if an error occured
 */
int timeutil_get_timeval_from_msec(int msec, struct timeval *tv)  
{
	int retval = 0;
	
	if (tv != NULL) {
		tv->tv_sec  = msec / 1000;
		tv->tv_usec = (msec % 1000) * 1000;
	}
	else {
		retval = -1;
	}

	return retval;
}

/**
 * Checks if a time interval is exceeded according to a given start time.
 * @param start     start time
 * @param interval  interval time
 * @return 1 if time interval is exceeded, else 0
 */
int timeutil_interval_exceeded(struct timeval *start, struct timeval *interval)
{
	int retval = 0;
	struct timeval diff;
	struct timeval now;

	if (start != NULL && interval != NULL) {
		gettimeofday(&now, NULL);
		timersub(&now, start, &diff);
		if (timercmp(&diff, interval, > )) {
			retval = 1;
		}
	}

	return retval;
}

/**
 * Calculates the length of time until the first of two given intervals gets
 * exceeded. If both intervals are exceeded a time value of 0 is returned.
 * \code
 * Example:
 *                interval1
 * start1 |------------------------|
 *                    interval2
 *     start2 |-----------------------|
 *                           diff
 *                    now |--------|
 * \endcode
 * @param start1     start time 1
 * @param interval1  interval time 1
 * @param start2     start time 2
 * @param interval2  interval time 2
 * @param diff       time until the first of two given intervals gets exceeded
 *                   (return value)
 * @return 0 on success or -1 if an error occured
 */
int timeutil_intervals_runtime(struct timeval *start1,
		struct timeval *interval1, struct timeval *start2,
		struct timeval *interval2, struct timeval *diff)
{
	int retval = -1;
	struct timeval now;
	struct timeval end1;
	struct timeval end2;

	if (start1 != NULL && interval1 != NULL && start2 != NULL
			&& interval2 != NULL && diff != NULL) {
		gettimeofday(&now, NULL);
		timeradd(start1, interval1, &end1);
		timeradd(start2, interval2, &end2);
		gettimeofday(&now, NULL);
		if (timercmp(&end1, &end2, > )) {
			timersub(&end2, &now, diff);
		}
		else {
			timersub(&end1, &now, diff);
		}
		if (diff->tv_sec < 0) {
			diff->tv_sec = 0;
			diff->tv_usec = 0;
		}
		retval = 0;
	}

	return retval;	
}
