/*
 * veronica.c - VNC server for Hifidelio devices. 
 *
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/** 
 * \mainpage
 * \image html hifidelio.jpg
 */

/**
 * \file veronica.c
 * VNC server for Hifidelio devices.
 */

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <rfb/rfb.h>
#include <pwd.h>
#include "config.h"
#include "veronica.h"

/**
 * User name of Veronica user.
 */
#define VERONICA_USER "fidelio"

/** 
 * Print usage text and command line options.
 */
void usage() 
{
	printf("\nUsage: veronica [OPTION]...\n");
	printf("\nOptions:");
	printf("\n  -d, --daemon            start as daemon process"); 
	printf("\n  -q, --quiet             no logging messages");
	printf("\n  -c, --config-file=FILE  configuration file; default is '/etc/veronica_conf.xml'");
	printf("\n  -h, --help              display this help message\n\n");
}

/**
 * Program entrance. Multiple things happen here:\n
 * - switch to 'fidelio' user id
 * - parsing of command line arguments
 * - daemon mode
 * - logging initialization
 * - VNC server startup
 *
 * @param argc  command line arguments count
 * @param argv  command line arguments
 * @returns process exit status (EXIT_SUCCESS, EXIT_FAILURE)
 */
int main(int argc, char** argv)
{
	int c, option_index = 0;
	int illegal_option  = 0;
	int logging   = 1;
	int daemonize = 0;
	char *config_file = NULL;
	struct config *config = NULL;
	struct passwd *veronica_user;

	/* Change user id */

    veronica_user = getpwnam(VERONICA_USER);
	if (veronica_user == NULL) {
		perror("main: getpwnam");
		goto error;
	}
	if (setreuid(veronica_user->pw_uid, veronica_user->pw_uid) != 0) {
		perror("main: setuid");
		goto error;
	}
	
	/* Parse command line arguments */

	static struct option long_options[] = {
		{"daemon",      no_argument,       0, 'd'},
		{"quiet",       no_argument,       0, 'q'},
		{"config-file", required_argument, 0, 'c'},
		{"help",        no_argument,       0, 'h'},
		{0, 0, 0, 0}
	};
	while (!illegal_option && (c = getopt_long (argc, argv, "dqc:h", 
			long_options, &option_index)) != -1) {
		switch (c) {
			case 'd':
				daemonize = 1;
				break;
			case 'q':
				logging = 0;
				break;
			case 'c':
				if ((config_file = strdup(optarg)) == NULL) {
					perror("main: strdup");
					goto error;
				}
				break;
			case 'h':
				usage();
				exit (EXIT_SUCCESS);
				break;
			case '?':
				illegal_option = 1;
				break;
			default:
				break;
		}
	}
	if (illegal_option) {
		fprintf(stderr, "Try 'veronica --help' for more information.\n");
		goto error;
	}

	/* Daemon mode and logging */

	if (daemonize) {
		if (daemon(1, logging) < 0) {
			perror("main: daemon");
			goto error;
		}
	} 
	rfbLogEnable(logging);
	
	/* VNC server configuration and startup */

	rfbLog("Veronica version %s starting...\n", VERONICA_VERSION);

	if ((config = config_load(config_file)) == NULL) {
		goto error;
	}
	vnc_startup(argc, argv, config);
	config_cleanup(config);
	free(config_file);
	exit(EXIT_SUCCESS);
	
error:	
	free(config_file);
	exit(EXIT_FAILURE);
}
