/*
 * vnc.c - Manages VNC server and screens.
 *
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/**
 * \file vnc.c
 * Manages VNC server and screens. Accomplishes the following tasks:\n
 * - VNC server/screen initialization
 * - screen refresh and cover lookup
 * - handling of key and mouse events
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <pthread.h>
#include <rfb/rfb.h>
#include "scale.h"
#include "timeutil.h"
#include "execsh.h"
#include "lang.h"
#include "display.h"
#include "cover.h"
#include "vnc.h"
#include "img.h"
#include "keys.h"
#include "xmlint.h"
#include "theme.h"

/**
 * Java classes directory.
 */
#define VNC_JAVA_CLASSES_DIR "/opt/veronica/classes"

/**
 * Bitmask of left mouse button (see ::vnc_do_ptr).
 */
#define VNC_MOUSE_BT_LEFT 1

/**
 * Bitmask of middle mouse button (see ::vnc_do_ptr).
 */
#define VNC_MOUSE_BT_MIDDLE 2

/**
 * Bitmask of right mouse button (see ::vnc_do_ptr).
 */
#define VNC_MOUSE_BT_RIGHT 4

/**
 * Bitmask of mouse wheel up action (see ::vnc_do_ptr).
 */
#define VNC_MOUSE_WHEEL_UP 8

/**
 * Bitmask of mouse wheel down action (see ::vnc_do_ptr).
 */
#define VNC_MOUSE_WHEEL_DOWN 16

/**
 * Controls the 'halt' button behaviour. If the button is released
 * within #VNC_HALT_DELAY_SEC seconds the Hifidelio performs a standby.
 * Otherwise, a shutdown is performed.
 */
#define VNC_HALT_DELAY_SEC 2.0f

/**
 * Cover art lookup interval in milliseconds.
 */
#define VNC_COVER_LOOKUP_MSEC 1000

/**
 * VNC event handling interval in microseconds. 
 */
#define VNC_DEFER_UPDATE_USEC 10000

/**
 * Color depth of VNC screen (bytes per pixel)
 */
#define VNC_COLOR_BPP 4

/**
 * VNC cursor width (see #x_cursor).
 */
#define VNC_CURSOR_WIDTH 11

/**
 * VNC cursor height (see #x_cursor).
 */
#define VNC_CURSOR_HEIGHT 17

/**
 * VNC cursor ASCII map.
 */
static char x_cursor[]=
	"           "
	" x         "
	" xx        "
	" xxx       "
	" xxxx      "
	" xxxxx     "
	" xxxxxx    "
	" xxxxxxx   "
	" xxxxxxxx  "
	" xxxxxxxxx "
	" xxxxxx    "
	" xxxxxx    "
	" xx  xxx   "
	" x   xxx   "
	"      xxx  "
	"      xx   "
	"           ";

/**
 * Data of a VNC screen.
 */
struct screen_data {
	struct theme *theme;             /**< Theme data. */
	int screen_width;                /**< Width of VNC screen. */
	int screen_height;               /**< Height of VNC screen. */
	int *y_scale_display;            /**< Y scaling vector (display). */
	int *y_roff_display;             /**< Y scaling vector offsets from the right. */
	int *y_loff_display;             /**< Y scaling vector offsets from the left. */
	int *x_scale_display;            /**< X scaling vector (display). */
	int *x_roff_display;             /**< X scaling vector offsets from the right. */
	int *x_loff_display;             /**< X scaling vector offsets from the left. */
	int display_scaled_width;        /**< Scaled width of display. */
	int display_scaled_height;       /**< Scaled height of display. */
	int client_count;                /**< Number of VNC clients. */
	struct timeval key_repeat_delay; /**< Minimum time between 'key presses'. */
	int mouse_control;               /**< Flag for mouse control within display area */
	int display_refresh;             /**< Screen refresh flag. */
	int cover_refresh;               /**< Cover refresh flag. */
};

/**
 * Data of a VNC client.
 */
struct client_data {
	rfbBool down;            /**< Down state of key event. */
	rfbKeySym key;           /**< Keycode of key event. */
	struct timeval key_ts;   /**< Timestamp of key event. */
	int button_mask;         /**< Mouse button mask of pointer event. */
	struct timeval click_ts; /**< Timestamp of mouse click event. */	
	struct timeval wheel_ts; /**< Timestamp of mouse wheel event. */	
};

/**
 * Global flag that keeps the VNC server running.
 */
static int keep_going = 1;

/**
 * Last signal handled in ::vnc_termination_handler.
 */
static int signal_caught = 0;

/**
 * Hifidelio screensaver mode (1 = active, 0 = off). This value gets updated
 * with every display refresh.
 */
static int screensaver_mode = 0;

/**
 * Called if a client quits the connection.
 * @param client_ctx  client context
 */
static void vnc_client_gone(rfbClientPtr client_ctx)
{
	rfbScreenInfoPtr screen = client_ctx->screen;
	struct screen_data *s_data = (struct screen_data *)screen->screenData;

	s_data->client_count--;
	free(client_ctx->clientData);
}

/**
 * Called if a new client connects to the VNC server.
 * @param client_ctx  client context
 * @return an action code whether the client is accepted
 */
static enum rfbNewClientAction vnc_new_client(rfbClientPtr client_ctx)
{
	rfbScreenInfoPtr screen = client_ctx->screen;
	struct screen_data *s_data = (struct screen_data *)screen->screenData;	
	struct client_data *c_data = 
			(struct client_data *)malloc(sizeof(struct client_data));

	if (c_data == NULL) {
		rfbLogPerror("vnc_new_client: malloc");
		return RFB_CLIENT_REFUSE;
	} 

	memset(c_data, 0, sizeof(struct client_data));
	client_ctx->clientData = c_data;
	client_ctx->clientGoneHook = vnc_client_gone;
	s_data->client_count++;
	s_data->display_refresh = 1;
	s_data->cover_refresh   = 1;
	
	return RFB_CLIENT_ACCEPT;
}

/**
 * Determines whether the pointer touches a button region.
 * @param button  button data
 * @param x       x-position of pointer
 * @param y       y-position of pointer
 * @return 1 if the pointer touches the button region, else 0
 */
static int vnc_check_ptr_over_btn_region(struct theme_button *button,
		int x, int y)
{
	int x1 = button->x;
	int y1 = button->y;
	int x2 = button->x + button->width;
	int y2 = button->y + button->height;

	return (x >= x1 && x <= x2 && y >= y1 && y <= y2);
}

/**
 * Determines whether the pointer touches a given region.
 * @param x1  upper left  x-coordinate of region
 * @param y1  upper left  y-coordinate of region
 * @param x2  lower right x-coordinate of region
 * @param y2  lower right y-coordinate of region
 * @param x   x-position of pointer
 * @param y   y-position of pointer
 * @return 1 if the pointer touches the region, else 0
 */
static int vnc_check_ptr_over_region(int x1, int y1, int x2, int y2,
		int x, int y)
{
	return (x >= x1 && x <= x2 && y >= y1 && y <= y2);
}

/**
 * Determines whether the pointer touches a transparent region in a bitmap.
 * @param bitmap   bitmap data
 * @param x        x-position of pointer
 * @param y        y-position of pointer
 * @return 1 if the region is transparent, else 0
 */
static int vnc_check_ptr_transparency(struct img_bitmap *bitmap, int x, int y)
{
	unsigned long *bm_ptr = (unsigned long *)bitmap->pixels;
	int offset = y * bitmap->width + x;

	return (*(bm_ptr + offset) == bitmap->transparency);
}

/**
 * Checks if the Hifidelio screensaver is active and sends a dummy keypress
 * if necessary.
 */
static void vnc_check_screensaver()
{
	if (screensaver_mode) {
		xmlint_send_key_press(SDLK_SCROLLOCK, KMOD_NONE, 1);
		xmlint_send_key_press(SDLK_SCROLLOCK, KMOD_NONE, 0);
	}
}

/**
 * Blits a button in the VNC screen's frame buffer. If the button is active,
 * the buttons bitmap is used as source. Otherwise, the layout bitmap is used.
 * @param theme   theme data
 * @param button  button data
 * @param screen  screen context
 */
static void vnc_blit_button(struct theme *theme, struct theme_button *button, 
		rfbScreenInfoPtr screen) 
{
	int x, y, offset;
	struct img_bitmap *bitmap;
	unsigned long *bm_ptr;
	unsigned long *fb_ptr;

	if (theme->buttons_bitmap == NULL) {
		return;
	}

	if (button->active) {
		bitmap = theme->buttons_bitmap;
	}
	else {
		bitmap = theme->layout_bitmap;
	}

	offset = button->y * bitmap->width + button->x;
	bm_ptr = ((unsigned long *)bitmap->pixels) + offset;
	fb_ptr = ((unsigned long *)screen->frameBuffer) + offset;

	for (y = 0; y < button->height; y++) {
		for (x = 0; x < button->width; x++) {
			if (*bm_ptr != bitmap->transparency) {
				*fb_ptr = *bm_ptr;
			}
			fb_ptr++;
			bm_ptr++;
		}
		fb_ptr += bitmap->width - x;
		bm_ptr += bitmap->width - x;
	}
}

/**
 * Checks the mouse button events for the given button and sends commands to
 * the XML interface according to the user's actions.
 * @param button       button data
 * @param button_mask  current mouse button mask
 * @param client_ctx   client context
 */
static void vnc_check_button_events(struct theme_button *button, 
		int button_mask, rfbClientPtr client_ctx)
{
	int i, keycode;
	char *message, *lang_resource;
	time_t now;
	rfbScreenInfoPtr screen = client_ctx->screen;
	struct client_data *c_data = (struct client_data *)client_ctx->clientData;

	if ((button_mask & VNC_MOUSE_BT_LEFT) 
			&& !(c_data->button_mask & VNC_MOUSE_BT_LEFT)) {
		vnc_check_screensaver();
		if (button->keycode != SDLK_HALT && button->keycode != SDLK_UNKNOWN) {
			xmlint_send_key_press(button->keycode, button->modifier, 1);
		}
		if (button->exec != NULL) {
			lang_resource = lang_get_resource(LANG_RES_BUTTON_EXEC);
			if (lang_resource != NULL) {
				message = (char *)malloc(strlen(lang_resource)
						+ strlen(button->name));
				if (message == NULL) {
					rfbLogPerror("vnc_check_button_events: malloc");
				}
				else {
					sprintf(message, lang_resource, button->name);
					display_show_message(message, 3000);
					execsh_exec(button->exec);
					free(message);
				}
			}
		}
		gettimeofday(&(c_data->click_ts), NULL);
	}
	else if (!(button_mask & VNC_MOUSE_BT_LEFT) 
			&& (c_data->button_mask & VNC_MOUSE_BT_LEFT)) {
		if (button->keycode == SDLK_HALT) {
			time(&now);
			if (difftime(now, c_data->click_ts.tv_sec) > VNC_HALT_DELAY_SEC) {
				keycode = SDLK_HALT;
				rfbLog("Power button released after standby period (halt)\n");
			} 
			else {
				keycode = SDLK_SLEEP;
				rfbLog("Power button released within standby period\n");
			}
			xmlint_send_key_press(keycode, KMOD_NONE, 1);
			xmlint_send_key_press(keycode, KMOD_NONE, 0);
		} 
		else if (button->keycode != SDLK_UNKNOWN) { 
			xmlint_send_key_press(button->keycode, button->modifier, 0);
		}
		gettimeofday(&(c_data->click_ts), NULL);
	}
}

/**
 * Checks mouse button events if the pointer doesn't touch a button region
 * in the current theme. Pressing the left mouse button acts as cursor-left
 * and pressing the right button as cursor-right in this case.
 * @param button_mask  current mouse button mask
 * @param client_ctx   client context
 */
static void vnc_check_nonbutton_events(int button_mask, rfbClientPtr client_ctx)
{
	rfbScreenInfoPtr screen = client_ctx->screen;
	struct client_data *c_data = (struct client_data *)client_ctx->clientData;

	if ((button_mask & VNC_MOUSE_BT_LEFT) 
			&& !(c_data->button_mask & VNC_MOUSE_BT_LEFT)) {
		vnc_check_screensaver();
		xmlint_send_key_press(SDLK_LEFT, KMOD_NONE, 1);
	}
	else if (!(button_mask & VNC_MOUSE_BT_LEFT) 
			&& (c_data->button_mask & VNC_MOUSE_BT_LEFT)) {
		xmlint_send_key_press(SDLK_LEFT, KMOD_NONE, 0);
	}
	else if ((button_mask & VNC_MOUSE_BT_RIGHT) 
			&& !(c_data->button_mask & VNC_MOUSE_BT_RIGHT)) {
		vnc_check_screensaver();
		xmlint_send_key_press(SDLK_RIGHT, KMOD_NONE, 1);
	}
	else if (!(button_mask & VNC_MOUSE_BT_RIGHT) 
			&& (c_data->button_mask & VNC_MOUSE_BT_RIGHT)) {
		xmlint_send_key_press(SDLK_RIGHT, KMOD_NONE, 0);
	}	
}

/**
 * Checks mouse wheel events and generates key-events (cursor-up,
 * cursor-down) if necessary.
 * @param button_mask  current mouse button mask
 * @param client_ctx   client context
 */
static void vnc_check_wheel_events(int button_mask, rfbClientPtr client_ctx)
{
	int keycode = -1;
	rfbScreenInfoPtr screen = client_ctx->screen;
	struct client_data *c_data = (struct client_data *)client_ctx->clientData;
	struct screen_data *s_data = (struct screen_data *)screen->screenData;

	if ((button_mask & VNC_MOUSE_WHEEL_UP) 
			&& !(c_data->button_mask & VNC_MOUSE_WHEEL_UP)) {
		keycode = SDLK_UP;
	}
	else if ((button_mask & VNC_MOUSE_WHEEL_DOWN) 
			&& !(c_data->button_mask & VNC_MOUSE_WHEEL_DOWN)) {
		keycode = SDLK_DOWN;
	}

	if (keycode != -1 && timeutil_interval_exceeded(&(c_data->wheel_ts), 
			&(s_data->key_repeat_delay))) {
		vnc_check_screensaver();
		xmlint_send_key_press(keycode, KMOD_NONE, 1);
		xmlint_send_key_press(keycode, KMOD_NONE, 0);
		gettimeofday(&(c_data->wheel_ts), NULL);
	}
}

/**
 * Handles pointer events.
 * @param button_mask  mouse button mask
 * @param x            x-coordinate of pointer
 * @param y            y-coordinate of pointer
 * @param client_ctx   client context
 */
static void vnc_do_ptr(int button_mask, int x, int y, rfbClientPtr client_ctx)
{
	int i, ptr_over, invalidate, button_region = 0;
	struct theme_button **buttons;
	rfbScreenInfoPtr screen = client_ctx->screen;
	struct client_data *c_data = (struct client_data *)client_ctx->clientData;
	struct screen_data *s_data = (struct screen_data *)screen->screenData;
	struct theme *theme = s_data->theme;

	if (theme == NULL) {
		return;
	}
	if (x < 0 || y < 0 
			|| x >= s_data->screen_width || y >= s_data->screen_height) {
		return;
	}

	buttons = theme->buttons;
	
	for (i = 0; i < theme->button_count; i++) {
		if (buttons[i]->width == 0 || buttons[i]->height == 0) {
			continue;
		}
		ptr_over = vnc_check_ptr_over_btn_region(buttons[i], x, y);
		if (ptr_over) {
			vnc_check_button_events(buttons[i], button_mask, client_ctx);
			button_region = 1;
		}
		if (ptr_over 
				&& !vnc_check_ptr_transparency(theme->buttons_bitmap, x, y)) {
			if (!buttons[i]->active) {
				buttons[i]->active = 1;
				vnc_blit_button(theme, buttons[i], screen);
				invalidate = 1;
			}
		}
		else if (buttons[i]->active) {
			buttons[i]->active = 0;
			vnc_blit_button(theme, buttons[i], screen);
			invalidate = 1;
		}
		else {
			invalidate = 0;
		}
		if (invalidate) {
			rfbMarkRectAsModified(screen, buttons[i]->x, buttons[i]->y, 
					buttons[i]->x + buttons[i]->width, 
					buttons[i]->y + buttons[i]->height);
		}
	}
	if (!button_region && s_data->mouse_control && theme->display != NULL
			&& vnc_check_ptr_over_region(theme->display->x, theme->display->y, 
					theme->display->x + s_data->display_scaled_width, 
					theme->display->y + s_data->display_scaled_height, x, y)) {
		vnc_check_nonbutton_events(button_mask, client_ctx);
		vnc_check_wheel_events(button_mask, client_ctx);
	}
	c_data->button_mask = button_mask;

	rfbDefaultPtrAddEvent(button_mask, x, y, client_ctx);
}

/**
 * Handles keyboard events and sends commands to the XML interface according
 * to the user's actions.
 * @param down        key state flag; 1: pressed, 0: released
 * @param key         keyboard code
 * @param client_ctx  client context
 */
static void vnc_do_key(rfbBool down, rfbKeySym key, rfbClientPtr client_ctx)
{
	rfbKeySym xk_modifier;
	rfbKeySym hf_keycode;
	rfbKeySym hf_modifier;
	int i;
	char *message, *lang_resource;
	struct theme_key **keys;	
	rfbScreenInfoPtr screen = client_ctx->screen;	
	struct client_data *c_data = (struct client_data *)client_ctx->clientData;
	struct screen_data *s_data = (struct screen_data *)screen->screenData;	
	struct theme *theme = s_data->theme;

	if (down && (timeutil_interval_exceeded(&(c_data->key_ts),
			&(s_data->key_repeat_delay)) || key != c_data->key)) {

		rfbLog("XK keycode: %d (0x%x)\n", key, key);

		if (keys_translate_vnc_hf(key, &hf_keycode, &hf_modifier) == 0) {
			vnc_check_screensaver();
			xmlint_send_key_press(hf_keycode, hf_modifier, 1);
			xmlint_send_key_press(hf_keycode, hf_modifier, 0);
		}
		if (theme != NULL && theme->key_count > 0) {
			xk_modifier = keys_is_xk_modifier(c_data->key) ? c_data->key : 0;
			keys = theme->keys;
			for (i = 0; i < theme->key_count; i++) {
				if (keys[i]->code == key && keys[i]->modifier == xk_modifier
						&& keys[i]->exec != NULL) {
					
					lang_resource = lang_get_resource(LANG_RES_KEY_EXEC);
					if (lang_resource != NULL) {						
						message = (char *)malloc(strlen(lang_resource)
								+ strlen(keys[i]->name));
						if (message == NULL) {
							rfbLogPerror("vnc_do_key: malloc");
							break;
						}
						else {
							sprintf(message, lang_resource, keys[i]->name);
							display_show_message(message, 3000);
							execsh_exec(keys[i]->exec);
							free(message);
						}
					}
				}
			}
		}
		c_data->down = down;
		c_data->key  = key;
		gettimeofday(&(c_data->key_ts), NULL);
	}
}

/**
 * Initializes a VNC screen. Performs the following tasks:\n
 * - loading of theme data
 * - framebuffer allocation
 * - display scaling preparations
 * - VNC screen initialization
 *
 * @param argc    command line argument count
 * @param argv    command line arguments
 * @param config  VNC configuration
 * @param index:  VNC screen index
 * @return pointer to an rfbScreenInfo structures or NULL if an error occured
 */
static rfbScreenInfoPtr vnc_init_screen(int argc, char** argv,
		struct config *config, int index) 
{
	int i, num_bytes;
	int screen_width, screen_height;
	char *theme_dir;
	char *auth_file    = NULL;
	void *frame_buffer = NULL;
	unsigned long *fb_ptr;
	struct theme_display *display;
	struct screen_data *s_data = NULL;
	struct theme *theme = NULL;
	rfbScreenInfoPtr screen;

	if (config->auth_file != NULL) {
		if ((auth_file = strdup(config->auth_file)) == NULL) {
			rfbLogPerror("vnc_init_screen: strdup");
			goto error;
		}
	}
	s_data = (struct screen_data *)malloc(sizeof(struct screen_data));
	if (s_data == NULL) {
		rfbLogPerror("vnc_init_screen: malloc");
		goto error;
	} 
	memset(s_data, 0, sizeof(struct screen_data));
	timeutil_get_timeval_from_msec(config->key_repeat_delay,
			&(s_data->key_repeat_delay));			
	theme_dir = config->screens[index]->theme_dir;
	if (theme_dir != NULL) {
		theme = theme_load(theme_dir);
		s_data->theme = theme;
	}
	s_data->display_refresh = 0;
	s_data->cover_refresh   = 0;
	s_data->mouse_control   = config->screens[index]->mouse_control;

	/* Framebuffer allocation */

	if (theme != NULL) {
		screen_width  = theme->layout_bitmap->width;
		screen_height = theme->layout_bitmap->height;
		num_bytes     = screen_width * screen_height * 4;
		frame_buffer  = (void *)malloc(num_bytes);
		if (frame_buffer == NULL) {
			rfbLogPerror("vnc_init_screen: malloc");
			goto error;
		}
		memcpy(frame_buffer, theme->layout_bitmap->pixels, num_bytes);
	}
	else {
		screen_width  = DISPLAY_WIDTH;
		screen_height = DISPLAY_HEIGHT;
		num_bytes     = screen_width * screen_height * 4;
		frame_buffer  = (void *)malloc(num_bytes);
		if (frame_buffer == NULL) {
			rfbLogPerror("vnc_init_screen: malloc");
			goto error;
		}
		fb_ptr = (unsigned long *)frame_buffer;
		for (i = 0; i < screen_width * screen_height; i++) {
			*(fb_ptr + i) = DISPLAY_BACK_COLOR;
		}
	}
	s_data->screen_width = screen_width;
	s_data->screen_height = screen_height;
	rfbLog("Size of framebuffer: %d x %d\n", screen_width, screen_height);

	/* Display scaling */

	if (theme != NULL && theme->display != NULL) {
		display = theme->display;
		s_data->display_scaled_width  = display->scale * DISPLAY_WIDTH; 
		s_data->display_scaled_height = display->scale * DISPLAY_HEIGHT;
	}
	else {
		s_data->display_scaled_width  = DISPLAY_WIDTH;
		s_data->display_scaled_height = DISPLAY_HEIGHT;
	}

	if (scale_vectors(DISPLAY_WIDTH,  s_data->display_scaled_width,
			&(s_data->x_scale_display), &(s_data->x_loff_display),
			&(s_data->x_roff_display)) == -1) {
		goto error;
	}
	if (scale_vectors(DISPLAY_HEIGHT, s_data->display_scaled_height,
			&(s_data->y_scale_display), &(s_data->y_loff_display),
			&(s_data->y_roff_display)) == -1) {
		goto error;
	}		
	
	rfbLog("Size of (possibly scaled) display: %d x %d\n",
			s_data->display_scaled_width, s_data->display_scaled_height);

	/* Screen initialization */

	screen = rfbGetScreen(&argc, argv, screen_width, screen_height, 
			8, 3, VNC_COLOR_BPP);
	if (screen == NULL) {
		rfbLog("ERROR: VNC screen initialization failed\n");
		goto error;
	}
	screen->kbdAddEvent    = vnc_do_key;
	screen->ptrAddEvent    = vnc_do_ptr;	
	screen->newClientHook  = vnc_new_client;
	screen->port           = config->screens[index]->vnc_port;
	screen->frameBuffer    = frame_buffer;
	screen->screenData     = s_data;
	screen->desktopName    = config->screens[index]->name;
	screen->authPasswdData = auth_file;
	if (config->screens[index]->http_port != CONFIG_NO_PORT) {
		screen->httpDir = VNC_JAVA_CLASSES_DIR;
		screen->httpEnableProxyConnect = TRUE;
	}
	if (config->screens[index]->show_cursor) { 
		screen->cursor = rfbMakeXCursor(VNC_CURSOR_WIDTH, VNC_CURSOR_HEIGHT,
				x_cursor, 0);
	}
	else {
		screen->cursor = NULL;
	}
	rfbInitServer(screen);
	rfbLog("Screen %p initialized successfully\n", screen);

	return screen;

error:
	free(frame_buffer);
	free(auth_file);
	if (s_data != NULL) {
		theme_cleanup(s_data->theme);
		free(s_data->x_scale_display);
		free(s_data->y_scale_display);
		free(s_data);
	}
	
	return NULL;
}

/**
 * Updates a display region in the screen's framebuffer.
 * @param screen          screen context
 * @param display_buffer  display buffer
 * @param x1              upper left  x-coordinate
 * @param y1              upper left  y-coordinate
 * @param x2              lower right x-coordinate
 * @param y2              lower right y-coordinate
 */
static void vnc_update_display_region(rfbScreenInfoPtr screen, 
		unsigned char *display_buffer, int x1, int y1, int x2, int y2)
{
	int x, y, i, offset, pixel;
	unsigned long fore_color;
	unsigned long *bm_ptr;
	unsigned long *fb_ptr = (unsigned long *)screen->frameBuffer;
	struct screen_data *s_data = (struct screen_data *)screen->screenData;
	struct theme *theme = s_data->theme;
	struct theme_display *display;
	int *y_scale = s_data->y_scale_display;
	int *x_scale = s_data->x_scale_display;
	int screen_width = s_data->screen_width;

	if (theme != NULL) {
		display = theme->display;
		x2 = display->width  < x2 ? display->width  : x2; 
		y2 = display->height < y2 ? display->height : y2; 
		if (x2 < x1 || y2 < y1) {
			return;
		}
	}

	x1 = s_data->x_loff_display[x1];
	x2 = s_data->x_roff_display[x2];
	y1 = s_data->y_loff_display[y1];
	y2 = s_data->y_roff_display[y2];

	if (theme != NULL) {
		bm_ptr = (unsigned long *)theme->layout_bitmap->pixels;
		fore_color = display->forecolor;
		offset = (display->y + y1) * screen_width + display->x + x1;
	} 
	else {
		fore_color = DISPLAY_FORE_COLOR;
		offset = y1 * screen_width + x1;
	}

	fb_ptr += offset;
	bm_ptr += offset;

	for (y = y1; y <= y2; y++) {
		offset = y_scale[y] * DISPLAY_WIDTH_BYTES;
		for (x = x1; x <= x2; x++) {
			pixel = display_buffer[offset + (x_scale[x] >> 3)]
				& (128 >> (x_scale[x] % 8));
			if (pixel == 0) {
				if (theme != NULL) {
					*fb_ptr = *bm_ptr;
				} 
				else {
					*fb_ptr = DISPLAY_BACK_COLOR;
				}
			}
			else {
				*fb_ptr = fore_color;
			}
			fb_ptr++;
			bm_ptr++;
		}
		bm_ptr += screen_width - (x - x1);
		fb_ptr += screen_width - (x - x1);
	}

	if (theme != NULL) {
		rfbMarkRectAsModified(screen, display->x + x1, display->y + y1,
				display-> x + x2 + 1, display->y + y2 + 1); 
	} 
	else {
		rfbMarkRectAsModified(screen, x1, y1, x2 + 1, y2 + 1); 
	} 
}

/**
 * Updates the cover region. Displays the background if no cover is available.
 * @param screen  screen context
 * @param cover   cover art bitmap
 */
static void vnc_update_cover(rfbScreenInfoPtr screen, struct img_bitmap *cover)
{
	int x, y, offset;
	int *x_scale = NULL, *y_scale = NULL;
	unsigned long *bm_ptr, *co_ptr, *fb_ptr;
	unsigned long c_pixel;
	unsigned int alpha1, alpha2, red, green, blue;
	struct screen_data *s_data = (struct screen_data *)screen->screenData;
	struct theme *theme = s_data->theme;

	if (theme == NULL || (theme != NULL && theme->cover == NULL)) {
		return;
	}
	
	offset = theme->cover->y * theme->layout_bitmap->width + theme->cover->x;
	fb_ptr = ((unsigned long *)screen->frameBuffer) + offset;
	bm_ptr = ((unsigned long *)theme->layout_bitmap->pixels) + offset;

	if (cover != NULL) {
		co_ptr = ((unsigned long *)cover->pixels);
		if (theme->cover->rotate == 90 || theme->cover->rotate == 270) {
			if (scale_vectors(cover->height, theme->cover->width,  &y_scale,
					NULL, NULL) == -1) {
				goto error;
			}
			if (scale_vectors(cover->width,  theme->cover->height, &x_scale,
					NULL, NULL) == -1) {
				goto error;
			}
		}
		else {
			if (scale_vectors(cover->width,  theme->cover->width,  &x_scale,
					NULL, NULL) == -1) {
				goto error;
			}
			if (scale_vectors(cover->height, theme->cover->height, &y_scale,
					NULL, NULL) == -1) {
				goto error;
			}
		}
		for (y = 0; y < theme->cover->height; y++) {
			for (x = 0; x < theme->cover->width; x++) {
				if (theme->cover->rotate == 90) {
					c_pixel = *(co_ptr + (y_scale[theme->cover->width - x - 1]
							* cover->width + x_scale[y]));
				}
				else if (theme->cover->rotate == 180) {
					c_pixel = *(co_ptr + (y_scale[theme->cover->height - y - 1]
							* cover->width
							+ x_scale[theme->cover->width - x - 1]));
				}
				else if (theme->cover->rotate == 270) {
					c_pixel = *(co_ptr + (y_scale[x] * cover->width
							+ x_scale[theme->cover->height - y - 1]));
				}
				else {
					c_pixel = *(co_ptr + (y_scale[y] * cover->width
							+ x_scale[x]));
				}
				if (cover->alpha_channel) {
					alpha1 = c_pixel & 0x000000ff;
					alpha2 = 0xff - alpha1;
					red   =  (alpha1 * ((c_pixel & 0xff000000) >> 24) 
							+ alpha2 * ((*bm_ptr & 0xff000000) >> 24)) >> 8;
					green =  (alpha1 * ((c_pixel & 0x00ff0000) >> 16) 
							+ alpha2 * ((*bm_ptr & 0x00ff0000) >> 16)) >> 8;
					blue  =  (alpha1 * ((c_pixel & 0x0000ff00) >> 8 ) 
							+ alpha2 * ((*bm_ptr & 0x0000ff00) >> 8 )) >> 8;
					*fb_ptr = 0L | (red << 24) | (green << 16) | (blue << 8);
				}
				else {
					*fb_ptr = c_pixel;
				}
				fb_ptr++;
				bm_ptr++;
			}
			fb_ptr += theme->layout_bitmap->width - x;
			bm_ptr += theme->layout_bitmap->width - x;
		}
	}
	else {
		for (y = 0; y < theme->cover->height; y++) {
			for (x = 0; x < theme->cover->width; x++) {
				*fb_ptr = *bm_ptr;
				fb_ptr++;
				bm_ptr++;
			}
			fb_ptr += theme->layout_bitmap->width - x;
			bm_ptr += theme->layout_bitmap->width - x;
		}		
	}
	rfbMarkRectAsModified(screen, theme->cover->x, theme->cover->y, 
			theme->cover->x + theme->cover->width, 
			theme->cover->y + theme->cover->height);

error:
	free(x_scale);
	free(y_scale);
}

/**
 * VNC screen cleanup. Frees allocated memory.
 * @param screen  VNC screen data
 */
static void vnc_cleanup_screen(rfbScreenInfoPtr screen) 
{
	struct screen_data *s_data;

	if (screen == NULL) {
		return;
	}
	
	if (screen->screenData != NULL) {
		s_data = (struct screen_data *)screen->screenData;
		theme_cleanup(s_data->theme);
		free(s_data->x_scale_display);
		free(s_data->y_scale_display);
		free(s_data);
	}
	free((char *)screen->desktopName);
	free(screen->authPasswdData);
	free(screen->frameBuffer);
	screen->frameBuffer = NULL;
	rfbScreenCleanup(screen);			
}

/**
 * Thread function that controls the event loop for a vnc screen.
 * @param arg  VNC screen data
 */
void *vnc_process_events(void *arg) {
	int i;
	rfbScreenInfoPtr screen = (rfbScreenInfoPtr)arg;

	while(keep_going) {
		rfbProcessEvents(screen, VNC_DEFER_UPDATE_USEC);
	}
	rfbLog("Event thread of screen %p finished\n", screen);
}

/**
 * Program termination signal handler.
 * @param sig  signal caught
 */
static void vnc_termination_handler(int sig)
{
	keep_going = 0;
	signal_caught = sig;
}

/**
 * Launches VNC screens and controls the refresh loop.
 * @param argc    command line argument count
 * @param argv    command line arguments
 * @param config  VNC configuration data
 */
void vnc_startup(int argc, char** argv, struct config *config)
{
	int i, j;
	int update_display, update_cover;
	int cover_status, cover_refresh;
	struct timeval cover_lookup_interval, cover_lookup_ts;
	struct timeval display_refresh_interval, display_refresh_ts;
	struct timeval idle_time;
	struct display_data *d_data;
	struct display_region *update_region;
	struct screen_data *s_data;
	struct theme *theme;	
	struct img_bitmap *cover = NULL, *new_cover;	
	static int old_cover_status = COVER_NONE;	
	rfbScreenInfoPtr *screens;
	pthread_t *process_events;


	/* Termination signal handling */

	signal (SIGTERM, vnc_termination_handler);
	signal (SIGINT,  vnc_termination_handler);
	signal (SIGQUIT, vnc_termination_handler);
	signal (SIGKILL, vnc_termination_handler);
	signal (SIGHUP,  vnc_termination_handler);	

	/* Timer initialization */
	
	timeutil_get_timeval_from_msec(config->refresh,
			&display_refresh_interval);
	timeutil_get_timeval_from_msec(VNC_COVER_LOOKUP_MSEC,
			&cover_lookup_interval);
	gettimeofday(&display_refresh_ts, NULL);			
	gettimeofday(&cover_lookup_ts, NULL);			
			
	/* Display and screen initialization */

	if ((d_data = display_get_data()) == NULL) {
		goto error;
	}

	screens = (rfbScreenInfoPtr *)malloc(sizeof(rfbScreenInfoPtr) 
			* config->screen_count);
	if (screens == NULL) {
		rfbLogPerror("vnc_startup: malloc");
		goto error;
	}
	memset(screens, 0, config->screen_count);
	for (i = 0; i < config->screen_count; i++) {
		if ((screens[i] = vnc_init_screen(argc, argv, config, i)) == NULL) {
			goto error;
		}
	}

	/* Event thread initialization */

	process_events = (pthread_t *)malloc(sizeof(pthread_t)
			* config->screen_count);
	if (process_events == NULL) {
		rfbLogPerror("vnc_startup: malloc");
		goto error;
	}
	memset(process_events, 0, config->screen_count);
	for (i = 0; i < config->screen_count; i++) {
		if (pthread_create(&(process_events[i]), NULL, vnc_process_events,
				screens[i]) != 0) {
			rfbLog("vnc_startup: Unable to create event thread\n");
			goto error;
		}
	}

	/* Refresh loop */

	while (keep_going) {

		/* Check update necessity */

		update_display = 0;
		update_cover   = 0;
		for (i = 0; i < config->screen_count; i++) {
			s_data = (struct screen_data *)screens[i]->screenData;
			theme = s_data->theme;
			if (s_data->client_count > 0) {
				update_display = 1;
				if (theme != NULL) {
					if (theme->display == NULL) {
						update_display = 0;
					}
					if (theme->cover != NULL) {
						update_cover = 1;
					}
				}
			}
		}

		/* Update display */

		if (update_display && timeutil_interval_exceeded(&display_refresh_ts,
				&display_refresh_interval)) {
			gettimeofday(&display_refresh_ts, NULL);			
			if (display_update(d_data) < 0) {
				goto error;
			}
			if (screensaver_mode != (d_data->info.bits_per_pixel == 0)) {
				rfbLog("Screensaver mode: %d\n", !screensaver_mode);
			}
			screensaver_mode = (d_data->info.bits_per_pixel == 0);
			for (i = 0; i < config->screen_count; i++) {
				s_data = (struct screen_data *)screens[i]->screenData;
				theme = s_data->theme;
				if (!rfbIsActive(screens[i]) || s_data->client_count == 0
						|| (theme != NULL && theme->display == NULL)) {
					continue;
				}
				if (s_data->display_refresh) {
					rfbLog("Screen %p display refresh\n", screens[i]);
					vnc_update_display_region(
							screens[i], d_data->display_buffer,
							0, 0, DISPLAY_WIDTH - 1, DISPLAY_HEIGHT - 1);
					s_data->display_refresh = 0;
				}
				else {
					for (j = 0; j < d_data->update_regions_count; j++) {
						update_region = d_data->update_regions[j];
						vnc_update_display_region(
								screens[i], d_data->display_buffer,
								update_region->x1, update_region->y1,
								update_region->x2, update_region->y2);
					}
				}
			}
		}

		/* Update cover */

		if (update_cover && timeutil_interval_exceeded(&cover_lookup_ts, 
				&cover_lookup_interval)) {
			gettimeofday(&cover_lookup_ts, NULL);
			new_cover = NULL;
			cover_status  = cover_check(&new_cover);
			cover_refresh = !(cover_status == COVER_ERR 
					|| cover_status == COVER_OLD 
					|| (old_cover_status == cover_status
						&&  cover_status == COVER_NONE));
			if (cover_refresh) {
				old_cover_status = cover_status;
				img_free(cover);
				cover = new_cover;
			}			
			for (i = 0; i < config->screen_count; i++) {
				s_data = (struct screen_data *)screens[i]->screenData;
				if ((cover_refresh || s_data->cover_refresh)
						&& rfbIsActive(screens[i])
						&& s_data->client_count > 0) {
					if (s_data->cover_refresh) {
						rfbLog("Screen %p cover refresh\n", screens[i]);
					}
					vnc_update_cover(screens[i], cover);
					s_data->cover_refresh = 0;
				}
			}
		}

		/* Sleep until next refresh */

		if (!update_display) {
			usleep(VNC_COVER_LOOKUP_MSEC * 1000);
		}
		else {
			timeutil_intervals_runtime(&display_refresh_ts,
					&display_refresh_interval, &cover_lookup_ts,
					&cover_lookup_interval, &idle_time);
			usleep(idle_time.tv_sec * 1000000 + idle_time.tv_usec);
		}
	}

error:
	if (process_events != NULL) {
		for (i = 0; i < config->screen_count; i++) {
			if (process_events[i] != 0) {
				pthread_join(process_events[i], NULL);
			}
		}
		free(process_events);
	}
	display_cleanup(d_data);
	img_free(cover);
	for (i = 0; i < config->screen_count; i++) {
		vnc_cleanup_screen(screens[i]);
	}
	if (!keep_going) {
		rfbLog("Termination signal caught. VNC server shutdown...\n");
		signal(signal_caught, SIG_DFL);
		raise(signal_caught);		
	}	
}

