/*
 * xmlint.c - Hifidelio XML interface access. 
 *
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/**
 * \file xmlint.c
 * Hifidelio XML interface access. Creates a socket connection to the XML
 * interface and sends commands there.
 */

#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <rfb/rfb.h>
#include "xmlint.h"

/**
 * Local port of the XML interface.
 */
#define XML_INT_PORT 50000

/**
 * XML command that simulates a key press.
 */
#define XML_INT_KEYPRESS "<KeyPress key=\"%i\" mod=\"%i\" state=\"%i\" />"

/**
 * File descriptor of the socket connection to the XML interface.
 */
static int fd_xml_sock = -1;

/**
 * Opens a connection to the XML interface (Port 50000, localhost).
 * @return a valid file descriptor or -1 if an error occured.
 */
static int xmlint_open_socket()
{
	int fd_sock;
	struct sockaddr_in server;
	pthread_t control;

	fd_sock = socket(PF_INET, SOCK_STREAM, 0);
	if (fd_sock < 0) {
		rfbLogPerror("xmlint_open_socket: socket");
		return -1;
	}

	server.sin_family = AF_INET;
	server.sin_addr.s_addr = htonl(INADDR_ANY);
	server.sin_port = htons(XML_INT_PORT);
	if (connect(fd_sock, (struct sockaddr *)&server, sizeof(server)) < 0) {
		rfbLogPerror("xmlint_open_socket: connect");
		return -1;
	}

	return fd_sock;
}

/**
 * Prepares and sends a <KeyPress> event with the given arguments.
 * @param keycode   keyboard scancode
 * @param modifier  keyboard modifier
 * @param down      key state flag; 1 : pressed, 0 : released
 * @return 0 on success or -1 if an error occured
 */
int xmlint_send_key_press(int keycode, int modifier, int down)
{
	char buf[256];

	if (fd_xml_sock < 0) {
		fd_xml_sock = xmlint_open_socket();
		if (fd_xml_sock < 0) {
			rfbLog("ERROR: XML interface connection failed\n");
			return -1;
		}
		rfbLog("XML interface connection at localhost:%d\n", XML_INT_PORT);
	}

	sprintf(buf, XML_INT_KEYPRESS, keycode, modifier, down);
	if (write(fd_xml_sock, buf, strlen(buf)) < 0) {
		rfbLogPerror("xmlint_send_key_press: write");
		fd_xml_sock = -1;
		return -1;
	}
	rfbLog("XML command: %s\n", buf);

	return 0;
}

/**
 * XML interface cleanup. Closes an open socket to the XML interface.
 */
void xmlint_cleanup()
{
	if (fd_xml_sock > -1) {
		close(fd_xml_sock);
	}
}

